import openai
import traceback
import threading
import time
import os
import sys
import signal
import json
import tempfile
import abc
import requests
import hashlib
import pickle
from llm_vm.guided_completion import Completion
# we need to package-ify so this works
from llm_vm.completion.data_synthesis import DataSynthesis
import inspect


job_id = None # we want to be able to cancel a fine_tune if you kill the program

def exit_handler(signum, frame):

    if (None != job_id):
        print("cancelling fine-tune if applicable", file=sys.stderr)
        openai.FineTune.cancel(id=job_id)

    print("user interrupt, exiting", file=sys.stderr)
    sys.exit()


if threading.current_thread() == threading.main_thread():        
    signal.signal(signal.SIGINT, exit_handler)


def generate_hash(input_string):
    sha256_hash = hashlib.sha256()
    sha256_hash.update(str(input_string).encode('utf-8'))
    return int(sha256_hash.hexdigest(), 16) % 10**18

def asyncStart(foo):
    t = [None, None]
    def new_thread():
        if callable(foo):
            t[0] = foo()
        else:
            t[0] = lambda x: x
    t[1] = threading.Thread(target=new_thread)
    t[1].start()
    return t

def asyncAwait(t):
    t[1].join()
    return t[0]


class LocalEphemeral:

    def __init__(self):
        self.training_store = {}

    def get_data(self, c_id):
        self.init_if_null(c_id)
        return self.training_store[c_id]["data"]

    def load_data(self,file):
        #using pickle files right now, in the future we will have to use databases
        self.training_store = pickle.load(file)

    def store_data(self,file):
         #using pickle files right now, in the future we will have to use databases
        pickle.dump(self.training_store,file)

    def add_example(self, c_id, example):
        self.init_if_null(c_id)
        self.training_store[c_id]["data"] += [example]

    def set_training_in_progress(self, c_id, is_training):
        self.init_if_null(c_id)
        self.training_store[c_id]["is_training"] = is_training

    def get_training_in_progress_set_true(self, c_id):
        self.init_if_null(c_id)
        # TODO: this could cause a concurrency bug when distributed!
        self.training_store[c_id]['lock'].acquire()
        old_val = self.training_store[c_id]["is_training"]
        if not old_val:
            self.training_store[c_id]["is_training"] = True
        self.training_store[c_id]['lock'].release()
        return old_val

    def set_model(self, c_id, model_id):
        self.init_if_null(c_id)
        self.training_store[c_id]["model"] = model_id

    def get_model(self, c_id):
        self.init_if_null(c_id)
        return self.training_store[c_id]["model"]

    def init_if_null(self, c_id):
        if not c_id in self.training_store:
            self.training_store[c_id] = { "is_training": False,
                                          'lock' : threading.Lock(),
                                          "data": [],
                                          "model": None }



class Optimizer:
    @abc.abstractmethod
    def complete(self, stable_context, dynamic_prompt, **kwargs):
        """
        Runs a completion using the string stable_context+dynamic_prompt.  Returns an optional training closure to use if the
        caller decides that the completion was particularly good.

        This method first checks if a model exists for the stable_context. If it does, it uses the model to complete the prompt.
        It then checks if the number of training examples is less than the maximum allowable. If it is, or if a model wasn't
        previously found, it retrieves the best completion for the prompt using a larger model, adds a new datapoint for training,
        and potentially fine-tunes a new model using the updated data, storing the new model if successful.

        The function returns the best completion (either generated by the stored model or the larger model).

        This can not handle cases where either stable_context or dynamic_prompt are just whitespace!

        Parameters:
        ----------
        - stable_context (str): Stable contextual data to use as a basis for training.
        - dynamic_prompt (str): The dynamic data to generate a completion for and potentially add to training data.
        - **kwargs: Additional arguments to be passed to the `call_small` and `call_big` methods.

        Returns:
        ----------
        - completion (str): The best completion for the dynamic prompt, as generated by either the stored model or the larger model.
        """


class HostedOptimizer(Optimizer):
    def __init__(self, anarchy_key, openai_key, MIN_TRAIN_EXS=20, MAX_TRAIN_EXS = 2000):
        self.anarchy_key = anarchy_key
        self.openai_key = openai_key
        self.MIN_TRAIN_EXS = MIN_TRAIN_EXS
        self.MAX_TRAIN_EXS = MAX_TRAIN_EXS

    def complete(self, stable_context, dynamic_prompt, **kwargs):
        """
        TODO: Runs the optimizing completion process on anarchy's hosted server with persistence.

        Parameters:
        ----------
        - stable_context (str): Stable contextual data to use as a basis for training.
        - dynamic_prompt (str): The dynamic data to generate a completion for and potentially add to training data.
        - **kwargs: Additional arguments to be passed to the `call_small` and `call_big` methods.

        Returns:
        ----------
        - completion (str): The best completion for the dynamic prompt, as generated by either the stored model or the larger model.
        """
        url = "https://api.chat.dev/completion/optimizing"
        payload = {**kwargs,
                   'stable_context': stable_context,
                   'dynamic_prompt': dynamic_prompt,
                   'anarchy_key' : self.anarchy_key,
                   'openai_key' : self.openai_key,
                   'MIN_TRAIN_EXS' : self.MIN_TRAIN_EXS,
                   'MAX_TRAIN_EXS' : self.MAX_TRAIN_EXS
                   }
        headers = {'Authorization': f'Bearer {self.anarchy_key}'}

        print("Payload: ", payload, file=sys.stderr)
        try:
            response = requests.post(url, json=payload, headers=headers)
            response.raise_for_status()  # Raise an exception for 4XX and 5XX status codes
            return response.json()['completion']
        except requests.exceptions.RequestException as e:
            print("Error occurred:", e, file=sys.stderr)


class LocalOptimizer(Optimizer):
    def __init__(self, storage=LocalEphemeral(), MIN_TRAIN_EXS = 1, MAX_TRAIN_EXS = 2000, call_small = None , call_big = None , big_model = None, small_model = None, openai_key="", variance=0.87, num_examples=50):
        self.storage = storage
        self.MIN_TRAIN_EXS = MIN_TRAIN_EXS
        self.MAX_TRAIN_EXS = MAX_TRAIN_EXS
        self.call_small = call_small
        self.call_big = call_big
        self.big_model = big_model
        self.small_model = small_model
        self.openai_key = openai_key
        self.data_synthesizer = DataSynthesis(variance, num_examples)

    def complete(self, stable_context, dynamic_prompt, data_synthesis = False, finetune = False, regex = None, type = None, choices = None, **kwargs):

        openai.api_key = self.openai_key
        completion, train = self.complete_delay_train(stable_context, dynamic_prompt, run_data_synthesis=data_synthesis, regex = regex, choices = choices, type = type, **kwargs)
        if finetune:
            train()
        return completion

    def complete_delay_train(self, stable_context, dynamic_prompt, run_data_synthesis = False, min_examples_for_synthesis = 1 ,c_id = None, regex = None, choices = None, type = None, grammar_type = None, **kwargs):
        """
        Runs a completion using the string stable_context+dynamic_prompt.  Returns an optional training closure to use if the
        caller decides that the completion was particularly good.

        This method first checks if a model exists for the stable_context. If it does, it uses the model to complete the prompt.
        It then checks if the number of training examples is less than the maximum allowable. If it is, or if a model wasn't
        previously found, it retrieves the best completion for the prompt using a larger model, adds a new datapoint for training,
        and potentially fine-tunes a new model using the updated data, storing the new model if successful.

        The function returns the best completion (either generated by the stored model or the larger model), and a closure
        function that encapsulates the fine-tuning process for potential execution at a later time.

        Parameters:
        ----------
        - stable_context (str): Stable contextual data to use as a basis for training.
        - dynamic_prompt (str): The dynamic data to generate a completion for and potentially add to training data.
        - c_id (str): To be used if multiple users could share the same stable_contexts so that we don't leak data.  If its None, defaults to all possible context.
        - **kwargs: Additional arguments to be passed to the `call_small` and `call_big` methods.

        Returns:
        ----------
        - completion (str): The best completion for the dynamic prompt, as generated by either the stored model or the larger model.
        - succeed_train (function): A closure function that encapsulates the fine-tuning process, ready for
        execution at a later time.  If you pass it a completion, it will use that, otherwise it will use the completion from the "best" model.
        """
        assert dynamic_prompt.strip() != "" or stable_context.strip() != ""
        assert self.call_big is not None and self.call_small is not None and self.big_model is not None and self.small_model is not None
        if stable_context.strip() == "" :
            print("Running with an empty context", file=sys.stderr)

        prompt = (stable_context + dynamic_prompt).strip()
        c_id_repr = str({'stable_context' : stable_context,
                    'args' : kwargs,
                    'MIN_TRAIN_EXS' : self.MIN_TRAIN_EXS,
                    'MAX_TRAIN_EXS' : self.MAX_TRAIN_EXS,
                    'call_small' : str(self.small_model).split(' ')[0], # HACKS
                    'call_big' : str(self.big_model).split(' ')[0],
                    }) if c_id is None else c_id
        c_id = generate_hash(c_id_repr)
        completion = None
        # will return None if no constraint is specified
        completion_model = Completion.create(regex, type, choices, grammar_type)

        model = self.storage.get_model(c_id)
        # this gives us the model_id
        if model is not None and completion_model is None:
            print("Using the new model:", model, flush=True, file=sys.stderr)
            completion = self.call_small(prompt = dynamic_prompt.strip(), model=model, **kwargs)

        training_exs = self.storage.get_data(c_id)

        best_completion_promise = None
        succeed_train = None
        best_completion = None

        if len(training_exs) < self.MAX_TRAIN_EXS:
            def promiseCompletion():
                if completion_model is not None:
                    best_completion = completion_model.complete(prompt)
                else:
                    best_completion = self.call_big(prompt, **kwargs)

                def actual_train(use_completion = None):

                    train_completion = best_completion if use_completion is None else use_completion
                    new_datapoint = (dynamic_prompt.strip(), train_completion)
                    self.storage.add_example(c_id, new_datapoint)
                    small_model_filename = kwargs.get("small_model_filename", None)

                    if run_data_synthesis:
                        if len(self.storage.get_data(c_id)) < min_examples_for_synthesis:
                            print("Data synthesis is not available right now, need more examples in storage.", file=sys.stderr)
                        else:
                            for j in self.data_synthesizer.data_synthesis(self.call_big, prompt,best_completion,openai_key=self.openai_key, completion=completion_model, **kwargs):
                                self.storage.add_example(c_id, j)
                    training_exs = self.storage.get_data(c_id)
                    print(training_exs, file=sys.stderr)
                    print("Considering Fine-tuning", flush=True, file=sys.stderr)

                    if len(training_exs) >= self.MIN_TRAIN_EXS and not self.storage.get_training_in_progress_set_true(c_id):
                        print("Actually Fine-tuning", flush=True, file=sys.stderr)
                        print("Training examples:",str(len(training_exs)), file=sys.stderr)
                        asyncStart(self.small_model.finetune(training_exs,self,c_id,small_model_filename))
                return (best_completion, actual_train)

            best_completion_promise = asyncStart(promiseCompletion)

            if completion is None:
                # crazy story: succeed_train gets set before this anyway if it makes sense to set it!
                completion, succeed_train = asyncAwait(best_completion_promise)

            else:
                _, succeed_train = asyncAwait(best_completion_promise)



        def succeed_train_closure(use_completion = None):
            def promise():
                if succeed_train is not None:
                    return succeed_train(use_completion)
                if best_completion_promise is not None:
                    try:
                        return asyncAwait(best_completion)[1](use_completion)
                    except:
                        return
            return asyncStart(promise)

        return completion, succeed_train_closure




def create_jsonl_file(data_list):
    out = tempfile.TemporaryFile('w+')
    for a,b in data_list:
        out.write(json.dumps({'prompt': a, 'completion': b}) + "\n")
    out.seek(0)
    return out



from llm_vm.utils.typings_llm import *

# Retuns the tools, in a random order, and with randomised IDs.
def get_randomised_training_tools (
    generic_tools: ToolList,
    shuffled_by: int = 0,
    shuffle_by_modulo: int = 1e9
) -> ToolList:

    def __shuffled(tool_id: int) -> int:
        return int((tool_id + shuffled_by) % shuffle_by_modulo)

    shuffled_generic_tools: ToolList = [{"id": __shuffled(t["id"]), "description": t["description"]} for t in generic_tools]

    tools: ToolList = shuffled_generic_tools + [
        {
            "id": __shuffled(DefaultTrainingTools.TRAIN_SEND_EMAIL.value),
            "description": "Use this tool to send an email to someone"
        },
        {
            "id": __shuffled(DefaultTrainingTools.TRAIN_BOOK_TABLE.value),
            "description": "Useful to book a table or slot at a restaurant"
        },
        {
            "id": __shuffled(DefaultTrainingTools.TRAIN_CREATE_DOCUMENT.value),
            "description": "Create a document on google drive"
        },
        {
            "id": __shuffled(DefaultTrainingTools.TRAIN_SHARE_DOCUMENT.value),
            "description": "Useful to share an online document with someone. You need the email"
        },
        {
            "id": __shuffled(DefaultTrainingTools.TRAIN_FIND_RESTAURANT.value),
            "description": "Use this tool to find a specific bar, café or restaurant, or to search for places to eat around a certain location."
        },
        {
            "id": __shuffled(DefaultTrainingTools.TRAIN_NEWS.value),
            "description": "Use this tool to get news and current events"
        },
        {
            "id": __shuffled(DefaultTrainingTools.TRAIN_BOOK_APPOINTMENT.value),
            "description": "Use this tool to book an appointment, meeting or doctor, through a calendar."
        },
        {
            "id": __shuffled(DefaultTrainingTools.TRAIN_AVAILABLE_APPOINTMENT.value),
            "description": "Use this tool to get the available time slots for meetings, doctors, appointments, etc."
        },
        {
            "id": __shuffled(DefaultTrainingTools.TRAIN_ORDER_FOOD.value),
            "description": "Use this tool to order food and get it delivered."
        }
    ]

    return tools



from llm_vm.utils.typings_llm import *

answer_from_memory_data: AnswerInMemoryModel = {
    "openai_model": OpenAIModel.DAVINCI,
    "data": [
        {
            "question": "What's the time?",
            "answer": False
        },
        {
            "mem": [
                ("Whats day is it today?", "Today is Sunday, December 15th, 2027. The time is 11:33AM")
            ],
            "question": "What's the time?",
            "answer": True
        },
        {
            "question": "How are you feeling?",
            "answer": True
        },
        {
            "question": "What color is the sky?",
            "answer": True
        },
        {
            "question": "What is the temperature in Portland?",
            "answer": False
        },
        {
            "mem": [
               ("What is the weather in Portland?", "It is sunny, 22C, with clear skies. Theres 40% chance of rain after 18:00hs")
            ],
            "question": "What is the temperature in Portland?",
            "answer": True
        },
        {
            "question": "What day is it today?",
            "answer": False
        },
        {
            "question": "What's the name of this place?",
            "answer": False
        }
    ]
}



from typing import Tuple, List, Dict, TypedDict, Union, Any, Optional
from enum import Enum

class OpenAIModel(Enum):
    CURIE = "curie"
    CURIE_TEXT = "text-curie-001"
    FAST_DAVINCI = "code-cushman-002"
    DAVINCI = "davinci"
    DAVINCI_TEXT = "text-davinci-003"

class LLMCallType(Enum):
    OPENAI_COMPLETION = "openai-completion"
    OPENAI_CHAT = "openai-chat"
    # ... other models

class DecisionStep(Enum):
    SPLIT = "split_questions"
    INPUT = "guess_api_input"
    FROM_MEMORY = "use_memory"
    TOOL_PICKER = "tool_picker"

class DefaultTools(Enum):
    I_DONT_KNOW = -1
    ANSWER_FROM_MEMORY = 0

    WOLFRAM = 1
    DIRECTIONS = 2
    WEATHER = 3

    ## Always set this to the latest ID. This is an integer, not an enum!
    __LAST_ID__: int = 4

class DefaultTrainingTools(Enum):
    TRAIN_SEND_EMAIL = 4
    TRAIN_CREATE_DOCUMENT = 5
    TRAIN_SHARE_DOCUMENT = 6
    TRAIN_BOOK_TABLE = 7
    TRAIN_FIND_RESTAURANT = 8
    TRAIN_NEWS = 9
    TRAIN_ORDER_FOOD = 10
    TRAIN_BOOK_APPOINTMENT = 11
    TRAIN_AVAILABLE_APPOINTMENT = 12
    TRAIN_FIND_FLIGHT = 13

TupleList = List[Tuple[str, str]]

class ToolTypeArgs(TypedDict):
    url: str
    params: Optional[dict]
    json: Optional[dict]
    jsonParams: Optional[dict]
    auth: Optional[dict]
    cert: Optional[str]
    data: Optional[dict]

class SingleTool(TypedDict):
    description: str
    dynamic_params: dict
    id: int
    args: ToolTypeArgs
    ai_response_prompt: Union[str, None]

ToolList = List[SingleTool]
# ToolList = NewType("ToolList", List[SingleTool])
EnumeratedToolList = List[Tuple[int,SingleTool]]

ALL_TOOLS_MAP: Dict[str, Union[DefaultTools, DefaultTrainingTools]] = {
    "dont_know": DefaultTools.I_DONT_KNOW,
    "memory": DefaultTools.ANSWER_FROM_MEMORY,
    "weather": DefaultTools.WEATHER,
    "wolfram": DefaultTools.WOLFRAM,
    "directions": DefaultTools.DIRECTIONS,
    "send_email": DefaultTrainingTools.TRAIN_SEND_EMAIL,
    "create_doc": DefaultTrainingTools.TRAIN_CREATE_DOCUMENT,
    "share_doc": DefaultTrainingTools.TRAIN_SHARE_DOCUMENT,
    "book_table": DefaultTrainingTools.TRAIN_BOOK_TABLE,
    "find_restaurant": DefaultTrainingTools.TRAIN_FIND_RESTAURANT,
    "news": DefaultTrainingTools.TRAIN_NEWS,
    "order_food": DefaultTrainingTools.TRAIN_ORDER_FOOD,
    "book_appointment": DefaultTrainingTools.TRAIN_BOOK_APPOINTMENT,
    "check_appointment": DefaultTrainingTools.TRAIN_AVAILABLE_APPOINTMENT,
    "find_flight": DefaultTrainingTools.TRAIN_FIND_FLIGHT
}

class QuestionSplitModelJSONData(TypedDict):
    mem: List[List[str]]
    question: str
    answer: List[str]

class QuestionSplitModelData(TypedDict):
    mem: Union[TupleList, None]
    question: str
    answer: Union[List[str], None]
    tools: Union[ToolList, None]

class AnswerFromMemoryModelData(TypedDict):
    mem: Union[TupleList, None]
    question: str
    thought: Union[str, None]
    answer: Union[str, None]

class QuestionSplitInputModel(TypedDict):
    model_name: str
    openai_model: OpenAIModel
    data: List[QuestionSplitModelData]

class PromptModelEntry(TypedDict):
    prompt: str
    completion: str

class LLMCallParams(TypedDict):
    llm: LLMCallType
    model: Tuple[str, Union[str, None]]
    prompt: str
    temperature: float
    max_tokens: int
    stop: str

LLMCallReturnType = Tuple[str, float]

class LLMTrainingResult(TypedDict):
    elapsed_time_s: int
    model_name: str
    model_files: List[str]

class LLMTrainingResultMeta(TypedDict):
    OPENAI_KEY: str
    DATE: str
    TIMESTAMP: int

LLModels = Union[Dict[str, LLMTrainingResult], LLMTrainingResultMeta]


class ToolInputModelJSONData(TypedDict):
    mem: Optional[List[str]]
    description: str
    question: str
    answer: dict
    params: dict
class ToolInputModelData(TypedDict):
    mem: Optional[TupleList]
    description: str
    question: str
    answer: dict
    params: dict

class ToolInputModel(TypedDict):
    openai_model: OpenAIModel
    data: List[ToolInputModelData]

class ToolpickerInputModelJSONData(TypedDict):
    mem: Optional[List[str]]
    thought: Optional[str]
    question: str
    answer: str
class ToolpickerInputModelData(TypedDict):
    mem: Optional[TupleList]
    thought: Optional[str]
    question: str
    answer: int
    tools: Optional[ToolList]

class ToolpickerInputModel(TypedDict):
    openai_model: OpenAIModel
    data: List[ToolpickerInputModelData]

class AnswerInMemoryModelData(TypedDict):
    mem: Optional[TupleList]
    facts: Optional[TupleList]
    question: str
    answer: Optional[bool]

class AnswerInMemoryModel(TypedDict):
    openai_model: OpenAIModel
    data: List[AnswerInMemoryModelData]

class DebugCalls(TypedDict):
    response: str
    gpt_suggested_params: Any
    url: str
    raw_api_return: str

DebugCallList = List[DebugCalls]



import sys
from flask import request, Blueprint
import json
import os
import openai
from llm_vm.agents.REBEL import agent
from llm_vm.client import Client
from llm_vm.config import settings
# load optimizer for endpoint use
# optimizer = LocalOptimizer(MIN_TRAIN_EXS=2,openai_key=None)

client = Client( big_model=settings.big_model, small_model=settings.small_model)

print('optimizer loaded', file=sys.stderr)

bp = Blueprint('bp',__name__)

@bp.route('/', methods=['GET'])
def home():
    return '''home'''

@bp.route('/v1/complete', methods=['POST'])
def optimizing_complete():
    rebel_agent = agent.Agent("", [], verbose=1)
    data = json.loads(request.data)
    static_context = data["context"]
    dynamic_prompt = data["prompt"]
    data_synthesis = False
    finetune = False
    use_rebel_agent = False
    kwargs = {}
    if "openai_key" not in data.keys():
        return {"status":0, "resp":"No OpenAI key provided"}

    if "temperature" in data.keys():
        if type(data["temperature"]) != float and type(data["temperature"]) != int:
            return {"status":0, "resp":"Wrong Data Type for temperature"}
        else:
            kwargs.update({"temperature":data["temperature"]})

    if "stoptoken" in data.keys():

        if type(data["stoptoken"]) != str and type(data["stoptoken"]) != list:
            # stop can either be a string or array of strings
            return {"status":0, "resp":"Wrong Data Type for stop"}
        elif type(data["stoptoken"]) == list:
            if len(data["stoptoken"]) > 4:
                return {"status":0, "resp":"Too many stop tokens in array limit to 4 or less"}
            # check that every element in the list is a string
            for j in data["stoptoken"]:
                if type(j) != str:
                    return {"status":0, "resp":"Wrong Data Type for stop"}
            kwargs.update({"stop": data["stoptoken"]})
        else:
            kwargs.update({"stop":data["stoptoken"]})

    if "data_synthesis" in data.keys():
        if type(data["data_synthesis"])==bool:
            data_synthesis = data["data_synthesis"]
        else:
            return {"status":0, "resp":"Wrong Data Type for data_synthesis"}

    if "finetune" in data.keys():
        if type(data["finetune"])==bool:
            finetune = data["finetune"]
        else:
            return {"status":0, "resp":"Wrong Data Type for finetune"}

    if "tools" in data.keys():
        if type(data["tools"]) != list:
            return {"status":0, "resp":"Wrong data type for tools list"}
        else:
            tools=[]
            for i in data["tools"]:
                temp_tool_dict = {}
                temp_args_dict = {}
                temp_tool_dict.update({"description":i["description"]})
                temp_tool_dict.update({"dynamic_params":i["dynamic_params"]})
                temp_tool_dict.update({"method":i["method"]})
                temp_args_dict.update({"url":i["url"]})
                temp_args_dict.update({"params":{}})
                for j in i["static_params"].keys():
                    temp_args_dict["params"].update({j:i["static_params"][j]})
                for k in i["dynamic_params"].keys():
                    temp_args_dict["params"].update({k:"{"+k+"}"})
                temp_tool_dict.update({"args":temp_args_dict})
                tools.append(temp_tool_dict)
            rebel_agent.set_tools(tools)
            use_rebel_agent = True
    try:
        openai.api_key = data["openai_key"]
    except:
        return  {"status":0, "resp":"Issue with OpenAI key"}

    # optimizer.openai_key = openai.api_key
    agent.set_api_key(openai.api_key,"OPENAI_API_KEY")
    try:
        if not use_rebel_agent:
            completion = client.complete(static_context,dynamic_prompt,openai_key=openai.api_key, data_synthesis=data_synthesis,finetune = finetune, **kwargs)
        else:
            completion = rebel_agent.run(static_context+dynamic_prompt,[])[0]
    except Exception as e:
        return {"status":0, "resp": str(e)}

    # return {"completion":completion, "status": 200}
    return completion



import json
import os
import sys

# # Get the current file's directory to grab the python files with common functionality in the utils/ folder
# current_dir = os.path.dirname(os.path.abspath(__file__))
# grandparent_dir = os.path.dirname(os.path.dirname(os.path.dirname(current_dir)))
# utils_dir = os.path.join(grandparent_dir, 'utils/')
# sys.path.append(utils_dir)

from llm_vm.utils.labels import *
from llm_vm.utils.typings_llm import *
from llm_vm.agents.FLAT.agent_helper.utils import make_interaction
from datetime import datetime
from llm_vm.agents.FLAT.agent_helper.tools import GENERIC_TOOLS
from random import sample, shuffle


def __create_tool_tag(tool: SingleTool) -> str:
    return f'<{L_TOOL} id="{tool["id"]}" description="{tool["description"]}"/>'

__DEFAULT_TOOLS = "\n".join(
    [__create_tool_tag(tool)
     for tool in GENERIC_TOOLS]
)

def __tools_we_have_access_to(tools: str = __DEFAULT_TOOLS) -> str:
    return f"""<{L_TOOL_LIST}>{tools}</{L_TOOL_LIST}>"""


def __split_into_subquestions_instructions(extra: str = "") -> str:
    return f"""<{L_SPLIT_INSTRUCTIONS}>Look at the tools we have access to. If needed, split <{L_QUESTION}/> into subquestions, so each <{L_QUESTION}/> can be answered with one use of one tool. Make as few subquestions as possible, comma-separated, avoid duplicates, and return nothing else.{extra}</{L_SPLIT_INSTRUCTIONS}>"""

def make_tool_desc(tool: SingleTool):
    params = "{" + ", ".join(['"'+l + '": '+v for l, v in tool['dynamic_params'].items()]
                               )+"}" if tool['dynamic_params'] != "" else "{}"
    return f'''<{L_TOOL} id="{tool['id']}">
  <{L_DESCRIPTION}>{tool['description']}</{L_DESCRIPTION}>
  <{L_PARAMS}>{params}</{L_PARAMS}>
</{L_TOOL}>
'''

def generate_convo_history(memory: TupleList = [], facts: TupleList = []) -> str:
    if len(memory) + len(facts) == 0:
        return ""

    return (
        "\n".join([make_interaction(p, a, '', L_QUESTION, L_ANSWER, L_ANSWER_DATA, INTERACTION = L_INT_HUMAN) for p, a in memory]) + \
        "\n".join([make_interaction(p, a, '', L_QUESTION, L_ANSWER, L_ANSWER_DATA, INTERACTION = L_INT_AI) for p, a in facts])
    )

def prompt_for_instructions(instructions: str) -> str:
    if instructions and len(instructions):
        return f'''[{L_BOT_INSTRUCTIONS}]: To summarize the JSON, remember:\n {instructions}\n\n\n'''
    else:
        return ""
def prompt_for_answer(question: str) -> str:
    return f'''[{L_QUESTION}]: {question}\n[{L_ANSWER}]: ''';

## ONLY for training purposes!
def get_training_tool_subset (
    list_of_tools: ToolList,
    tool_id_always_returned: Union[int, None],
    max_num_elements: int = 6
) -> ToolList:
    # ONLY when we don't know, we return any elements.
    if tool_id_always_returned == None \
       or tool_id_always_returned == DefaultTools.I_DONT_KNOW.value \
       or tool_id_always_returned == DefaultTools.ANSWER_FROM_MEMORY.value:
        return sample(list_of_tools, max_num_elements - 1)

    specific_tool = next(tool for tool in list_of_tools if tool["id"] == tool_id_always_returned)
    subset_of_tools = sample(list_of_tools, max_num_elements - 1)
    # make sure it doesn't have our specific_tool inside
    subset_of_tools = [tool for tool in subset_of_tools if tool["id"] != tool_id_always_returned] + [specific_tool]

    # VERY IMPORTANT - otherwise the model thinks that the last answer is always the correct one.
    shuffle(subset_of_tools)

    return subset_of_tools


def splitter_prompt(elements: List[QuestionSplitModelData]) -> Tuple[str, str]:

    stopper: str = f'''</{L_ANSWER}>
</{L_SPLIT}>'''

    last_stop = None;

    response: str = f'''\
{__split_into_subquestions_instructions()}'''

    for element in elements:
        response += f'''\
<{L_SPLIT}>
{
    __tools_we_have_access_to("".join([__create_tool_tag(tool) for tool in element["tools"]])) if ("tools" in element and element["tools"] and isinstance(element["tools"], list)) else ""
}{
    f"""<{L_CONVERSATION}>{"".join([make_interaction(q,a) for (q,a) in element["mem"]])}</{L_CONVERSATION}>""" if ("mem" in element and isinstance(element["mem"], list)) else ""
}
<{L_QUESTION}>{element["question"]}</{L_QUESTION}>
<{L_ANSWER}>'''

        if "answer" in element and element["answer"]:
            response += f'''{SUBQ_SPLITTER.join(element["answer"])}{stopper}\n'''
            last_stop = None
        else:
            last_stop = stopper

    return response, last_stop


def toolpicker_prompt(elements: List[ToolpickerInputModelData], tools: ToolList) -> str:

    shuffled_tools = tools.copy()
    shuffle(shuffled_tools)
    tools_html = "\n".join([__create_tool_tag(_tool) for _tool in shuffled_tools])
    last_stop = None

    prompt: str = f'''<{L_BOT_INSTRUCTIONS}><{L_DESCRIPTION}>
We have a list of tools in <{L_TOOL_LIST}/>, and a question <{L_QUESTION}/> that we want to answer.
Which of the <{L_TOOL}/> in <{L_TOOL_LIST}/> can be used to best answer <{L_QUESTION}/>?
Answer only with the 'id' of the most relevant <{L_TOOL}/> (i.e.: {", ".join([str(t['id']) for t in shuffled_tools])}).
Answer {DefaultTools.ANSWER_FROM_MEMORY.value} if the answer is in <{L_CONVERSATION}/>.
Answer {DefaultTools.I_DONT_KNOW.value} if no tool can be used to answer <{L_QUESTION}/>.
</{L_DESCRIPTION}>

<{L_TOOL_LIST}>
{tools_html}
</{L_TOOL_LIST}>
</{L_BOT_INSTRUCTIONS}>
'''

    for element in elements:

        prompt += f'''<{L_INTERACTION}>'''

        prompt += f'''<{L_CONVERSATION}>
            {"".join([make_interaction(mem_q, mem_a) for mem_q, mem_a in (element["mem"])
        ])}</{L_CONVERSATION}>''' if "mem" in element and isinstance(element["mem"], list) and len(element["mem"]) else ""

        prompt += f'''{
      f'<{L_THOUGHT}>{element["thought"]}</{L_THOUGHT}>' if "thought" in element and element["thought"] else ""
}
<{L_QUESTION}>{element["question"]}</{L_QUESTION}>
<{L_ANSWER}>'''

        stopper = f'''</{L_ANSWER}>\n</{L_INTERACTION}>'''

        if "answer" in element:
            prompt += f'''{element["answer"]}{stopper}\n'''
            last_stop = None
        else:
            last_stop = stopper

    return prompt, last_stop


def create_memory_prompt(elements: List[AnswerInMemoryModelData]) -> Tuple[str, str]:
    do_you_know_the_answer = f"<{L_BOT_INSTRUCTIONS}>Is the answer to <{L_QUESTION}/> found in the memory or in your knowledge base already? Answer with a yes or no.</{L_BOT_INSTRUCTIONS}>"

    stop = f"</{L_ANSWER}></{L_INTERACTION}>"

    last_stop = None

    prompt: str = f'''{do_you_know_the_answer}\n'''

    for element in elements:
        conversation_history = generate_convo_history(
            element["mem"] if "mem" in element and element["mem"] else [],
            element["facts"] if "facts" in element and element["facts"] else []
        )
        prompt += f'''<{L_INTERACTION}>\n'''
        if conversation_history:
            prompt += f'''\
<{L_CONVERSATION}>
    {conversation_history}
</{L_CONVERSATION}>\n'''
        prompt += f'''<{L_QUESTION}>{element["question"]}</{L_QUESTION}>\n<{L_ANSWER}>'''

        if "answer" in element and isinstance(element["answer"], bool):
            prompt += f'''{"yes" if element["answer"] else "no"}{stop}'''
            last_stop = None
        else:
            last_stop = stop
            break

    return prompt, last_stop




def make_tool_input_case(
    q_facts: TupleList,
    q_question: str,
    answer_json_input: Union[dict, None],
    tool_descr: str,
    tool_params: dict,
    answer_label: str = "JSON",
    wrapper_tag: str = L_EXAMPLE,
) -> Tuple[str, Union[str, None]]:
    system_instructions: str = f'''
We have the following {answer_label}: <{answer_label}>{json.dumps(tool_params)}</{answer_label}>. Each key describes what its value is.\n
The {answer_label} is used where: <{L_DESCRIPTION}>{tool_descr}</{L_DESCRIPTION}>\n
We need a {answer_label} object with the same keys as <{answer_label}/> to answer <{L_QUESTION}/>{f""" (we can use the <{L_CONVERSATION}/> for contextual info as well)""" if len(q_facts) else ""}. Must be a valid {answer_label}!'''

    stopper: str = f"</{L_ANSWER}>\n</{wrapper_tag}>"

    mem = generate_convo_history(facts=q_facts)
    return f"\n<{wrapper_tag}>\n" \
        + f"  <{L_BOT_INSTRUCTIONS}>{system_instructions}</{L_BOT_INSTRUCTIONS}>" \
        + (f"  <{L_CONVERSATION}>{mem}\n</{L_CONVERSATION}>\n" if mem else "") \
        + f"  <{L_QUESTION}>{q_question}</{L_QUESTION}>\n" \
        + f"  <{L_ANSWER} type=\"{answer_label}\">"\
        + (( f"{json.dumps(answer_json_input)}{stopper}") if answer_json_input else ""), None if answer_json_input else stopper



from abc import ABC,abstractmethod
import sys
import openai
import math
from ctransformers import AutoModelForCausalLM
from transformers import (
    AutoModelForMaskedLM,
    AutoModelForSeq2SeqLM,
    AutoTokenizer,
    OPTForCausalLM,
    BloomForCausalLM,
    GPTNeoForCausalLM,
    GPTNeoXForCausalLM,
    LlamaForCausalLM,
    LlamaTokenizer,
    CodeLlamaTokenizer,
    DataCollatorForLanguageModeling,
    TrainingArguments,
    Trainer,
    BitsAndBytesConfig)
import time
from datetime import datetime
import tempfile
import json
import os
import torch
from peft import get_peft_model, LoraConfig, prepare_model_for_kbit_training
from trl import SFTTrainer
from sentence_transformers import SentenceTransformer



__private_key_value_models_map =  {}
# []   {
#         "opt": SmallLocalOpt,
#         "bloom": SmallLocalBloom,
#         "neo": SmallLocalNeo,
#         "llama2": SmallLocalLLama,
#         "gpt": GPT3,
#         "chat_gpt": ChatGPT,
#         "flan" : SmallLocalFlanT5,
#         "pythia" : SmallLocalPythia,
#         }

# Check available devices
if torch.cuda.device_count() > 1:
    device = [f"cuda:{i}" for i in range(torch.cuda.device_count())]  # List of available GPUs
else:  # If only one GPU is available, use cuda:0, else use CPU
    device = "cuda:0" if torch.cuda.is_available() else "cpu"

def RegisterModelClass(name):
    def regClass(cls):
        __private_key_value_models_map[name]=cls
    return regClass

model_keys_registered = __private_key_value_models_map.keys()
# Dictionary of models to be loaded in ModelConfig
def load_model_closure(model_name):
    models = __private_key_value_models_map
    return models[model_name]

# this is a hack till we add dynaconf or something?
if os.name == "nt":
    homepath = os.path.join('C:\\','Users',os.getlogin())
else:
    homepath = os.environ.get("HOME")

model_path_default = os.path.join( homepath , ".llm_vm", "models")
os.makedirs(model_path_default, exist_ok = True)

def create_jsonl_file(data_list):
    out = tempfile.TemporaryFile('w+')
    for a,b in data_list:
        out.write(json.dumps({'prompt': a, 'completion': b}) + "\n")
    out.seek(0)
    return out


class FinetuningDataset(torch.utils.data.Dataset):
    def __init__(self,iterable_dataset,length):
        self.dataset = list(iterable_dataset)
        self.length = length
    def __len__(self):
        return self.length
    def __getitem__(self, idx):
        return self.dataset[idx]

class BaseOnsiteLLM(ABC):
    def __init__(self,model_uri=None, tokenizer_kw_args={}, model_kw_args={}):
        if model_uri != None :
            self.model_uri= model_uri
        if model_uri is None and self.model_uri is None:
            raise ValueError('model_uri not found')
        self.model_name : str = self.model_uri.split('/')[-1] # our default for deriving model name
        self.model=self.model_loader(**model_kw_args)
        self.tokenizer=self.tokenizer_loader(**tokenizer_kw_args)

        # Move the model to the specified device(s)
        if isinstance(device, list):
            # If multiple GPUs are available, use DataParallel to parallelize the model
            self.model = torch.nn.DataParallel(self.model, device_ids=[i for i in range(len(device))])
            self.model.to(device[0])  # Move model to the first GPU in the list
            print(f"`{self.model_uri}` loaded on {len(device)} GPUs.", file=sys.stderr)
        else:
            self.model.to(device)  # Move model to the selected device (single GPU or CPU)
            print(f"`{self.model_uri}` loaded on {device}.", file=sys.stderr)

    @property
    @abstractmethod
    def model_uri(self):
        pass

    @model_uri.setter
    def model_uri(self,val):
        self.model_uri=val # check if this is correct

    # model_name : str = self.model_uri.split('/')[-1]

    @abstractmethod
    def model_loader(self):
        pass

    @abstractmethod
    def tokenizer_loader(self):
        pass

    def load_finetune(self, model_filename):
        self.model.load_state_dict(torch.load(os.path.join(model_path_default,"finetuned_models", self.model_name, model_filename)))


    def generate(self,prompt,max_length=100, tokenizer_kwargs={}, generation_kwargs={}): # both tokenizer and model take kwargs :(
        """
        This function uses the class's llm and tokenizer to generate a response given a user's prompt

        Parameters:
            prompt (str): Prompt to send to LLM
            max_length (int): Optional parameter limiting response length


        Returns:
            str: LLM Generated Response

        Example:
           >>> SmallLocalOpt.generate("How long does it take for an apple to grow?")
           I think it takes about a week for the apple to grow.
        """


        if isinstance(device, list):
            # If multiple GPUs are available, use first one
            inputs = self.tokenizer(prompt, return_tensors="pt", **tokenizer_kwargs).to(device[0])
        else:
            inputs = self.tokenizer(prompt, return_tensors="pt").to(device)
        generate_ids=self.model.generate(inputs.input_ids, max_length=max_length, **generation_kwargs)
        resp= self.tokenizer.batch_decode(generate_ids,skip_special_tokens=True,clean_up_tokenization_spaces=False)[0]
        # need to drop the len(prompt) prefix with these sequences generally
        # because they include the prompt.
        return resp[len(prompt):]

    def finetune(self,data, optimizer, c_id, model_filename=None):
        def asynctune():
            old_model = optimizer.storage.get_model(c_id)
            if old_model is not None:
                self.model.load_state_dict(torch.load(old_model))
            untokenized_final_dataset = []
            for prompt,response in data:
                untokenized_final_dataset.append(prompt + response)
            tokenized_final_dataset = map(self.tokenizer,untokenized_final_dataset)
            self.tokenizer.pad_token = self.tokenizer.eos_token
            data_collator = DataCollatorForLanguageModeling(tokenizer=self.tokenizer, mlm=False)
            optimizer.storage.set_training_in_progress(c_id, True)
            training_args = TrainingArguments(
                output_dir=os.path.join(model_path_default,"finetuned_models",),
                evaluation_strategy="epoch",
                learning_rate=2e-5,
                per_device_train_batch_size = 1,
                per_device_eval_batch_size = 1,
                num_train_epochs=5,
                weight_decay=0.01,
                report_to= "none",
            )
            test_set = FinetuningDataset(tokenized_final_dataset,len(untokenized_final_dataset))

            trainer = Trainer(
                model=self.model,
                args=training_args,
                train_dataset=test_set,
                eval_dataset=test_set,
                data_collator=data_collator,
            )
            os.makedirs(os.path.join(model_path_default,"finetuned_models", self.model_name), exist_ok=True)
            if tokenized_final_dataset:
                trainer.train()
                eval_results = trainer.evaluate()
            optimizer.storage.set_training_in_progress(c_id, False)

            if os.name == "nt":
                timestamp = datetime.now().strftime('%Y-%m-%dT%H-%M-%S')
            else:
                timestamp = datetime.now().strftime('%Y-%m-%dT%H:%M:%S')
            new_model = os.path.join(model_path_default,"finetuned_models",self.model_name, timestamp + '_' + self.model_name + ".pt" ) if model_filename is None else os.path.join(model_path_default,"finetuned_models",model_filename)
            open(new_model,"a")
            torch.save(self.model.state_dict(), new_model) # the model in memory is different now
            self.model_name = self.model_name + "_ft_"+  timestamp
            optimizer.storage.set_model(c_id, new_model)
            return math.exp(eval_results['eval_loss']) #perplexity is the metric we use for finetuning measurement
        return asynctune


    def lora_finetune(self, data, optimizer, c_id, model_filename=None):
        def async_lora():
            old_model = optimizer.storage.get_model(c_id)
            if old_model is not None:
                self.model.load_state_dict(torch.load(old_model))
            untokenized_final_dataset = []
            for prompt,response in data:
                untokenized_final_dataset.append(prompt + response)
            tokenized_final_dataset = map(self.tokenizer,untokenized_final_dataset)
            self.tokenizer.pad_token = self.tokenizer.eos_token
            data_collator = DataCollatorForLanguageModeling(tokenizer=self.tokenizer, mlm=False)
            optimizer.storage.set_training_in_progress(c_id, True)
            training_args = TrainingArguments(
                output_dir=os.path.join(model_path_default,"finetuned_models",),
                evaluation_strategy="epoch",
                learning_rate=2e-5,
                per_device_train_batch_size = 1,
                per_device_eval_batch_size = 1,
                num_train_epochs=5,
                weight_decay=0.01,
                report_to= "none",
            )
            test_set = FinetuningDataset(tokenized_final_dataset,len(untokenized_final_dataset))

            peft_config = LoraConfig(
                r=16,
                lora_alpha=32,
                lora_dropout=0.05,
                bias="none",
                task_type="CAUSAL_LM",
            )
            
            trainer = SFTTrainer(
                self.model,
                args=training_args,
                train_dataset=test_set,
                eval_dataset=test_set,
                data_collator=data_collator,
                peft_config=peft_config
            )
            os.makedirs(os.path.join(model_path_default,"finetuned_models", self.model_name), exist_ok=True)
            if tokenized_final_dataset:
                trainer.train()
                eval_results = trainer.evaluate()
            optimizer.storage.set_training_in_progress(c_id, False)

            if os.name == "nt":
                timestamp = datetime.now().strftime('%Y-%m-%dT%H-%M-%S')
            else:
                timestamp = datetime.now().strftime('%Y-%m-%dT%H:%M:%S')
            new_model = os.path.join(model_path_default,"finetuned_models",self.model_name, timestamp + '_' + self.model_name + ".pt" ) if model_filename is None else os.path.join(model_path_default,"finetuned_models",model_filename)
            open(new_model,"a")
            torch.save(self.model.state_dict(), new_model) # the model in memory is different now
            self.model_name = self.model_name + "_ft_"+  timestamp
            optimizer.storage.set_model(c_id, new_model)
            return math.exp(eval_results['eval_loss']) #perplexity is the metric we use for finetuning measurement
        return async_lora
    
    def quantize_model(self, bits=4):
        if self.model.is_quantizable():
            q4_config = BitsAndBytesConfig(load_in_4bit=True, bnb_4bit_use_double_quant=True, bnb_4bit_quant_type='nf4', bnb_4bit_compute_dtype=torch.bfloat16)
            q8_config = BitsAndBytesConfig(load_in_8bit=True, bnb_4bit_compute_dtype=torch.bfloat16)
            
            if bits == 4:
                q_model = AutoModelForCausalLM.from_pretrained(self.model_uri, quantization_config=q4_config)
            elif bits == 8:
                q_model = AutoModelForCausalLM.from_pretrained(self.model_uri, quantization_config=q8_config)
            else:
                raise ValueError("Only 4-bit and 8-bit quantization supported")
            return q_model
        else:
            raise NotImplementedError(f"{self.model} cannot be quantized")


    def qlora_finetune(self, data, optimizer, c_id, model_filename=None):
        def async_qlora():
            old_model = optimizer.storage.get_model(c_id)
            if old_model is not None:
                self.model.load_state_dict(torch.load(old_model))
            untokenized_final_dataset = []
            for prompt,response in data:
                untokenized_final_dataset.append(prompt + response)
            tokenized_final_dataset = map(self.tokenizer,untokenized_final_dataset)
            self.tokenizer.pad_token = self.tokenizer.eos_token
            data_collator = DataCollatorForLanguageModeling(tokenizer=self.tokenizer, mlm=False)
            optimizer.storage.set_training_in_progress(c_id, True)
            training_args = TrainingArguments(
                output_dir=os.path.join(model_path_default,"finetuned_models",),
                evaluation_strategy="epoch",
                per_device_train_batch_size=1,
                gradient_accumulation_steps=4,
                warmup_steps=2,
                max_steps=10,
                learning_rate=2e-4,
                fp16=True,
                logging_steps=1,
                optim="paged_adamw_8bit"
            )
            test_set = FinetuningDataset(tokenized_final_dataset,len(untokenized_final_dataset))
            
            self.model.gradient_checkpointing_enable()
            self.model = prepare_model_for_kbit_training(self.model)
            config = LoraConfig(
                r=8, 
                lora_alpha=32, 
                target_modules=["query_key_value"], 
                lora_dropout=0.05, 
                bias="none", 
                task_type="CAUSAL_LM"
            )
            self.model = get_peft_model(self.model, config)
            trainer = Trainer(
                self.model,
                args=training_args,
                train_dataset=test_set,
                eval_dataset=test_set,
                data_collator=data_collator,
            )
            os.makedirs(os.path.join(model_path_default,"finetuned_models", self.model_name), exist_ok=True)
            if tokenized_final_dataset:
                trainer.train()
                eval_results = trainer.evaluate()
            optimizer.storage.set_training_in_progress(c_id, False)

            if os.name == "nt":
                timestamp = datetime.now().strftime('%Y-%m-%dT%H-%M-%S')
            else:
                timestamp = datetime.now().strftime('%Y-%m-%dT%H:%M:%S')
            new_model = os.path.join(model_path_default,"finetuned_models",self.model_name, timestamp + '_' + self.model_name + ".pt" ) if model_filename is None else os.path.join(model_path_default,"finetuned_models",model_filename)
            open(new_model,"a")
            torch.save(self.model.state_dict(), new_model) # the model in memory is different now
            self.model_name = self.model_name + "_ft_"+  timestamp
            optimizer.storage.set_model(c_id, new_model)
            return math.exp(eval_results['eval_loss']) #perplexity is the metric we use for finetuning measurement
        return async_qlora
    
    def finetune_immediately(self):
        self.finetune()()

    def lora_finetune_immediately(self):
        self.lora_finetune()()

    def qlora_finetune_immediately(self):
        self.qlora_finetune()()



"""
this factorization isn't necessarily the greatest, nor should it be viewed
as likely being more general, aside from covering hugging face transformers
"""
@RegisterModelClass("pythia")
class SmallLocalPythia(BaseOnsiteLLM):
    """
    This is a class for ElutherAI's Pythia-70m LLM

    Attributes:
        model_uri (str): Hugging Face Endpoint for LLM
        tokenizer (AutoTokenizer): Tokenizer from Transformer's library
        model (LLM): The large language model

    Methods:
        model_loader: Loads the LLM into memory
        tokenizer_loader: Loads the tokenizer into memory
        generate: Generates a response from a given prompt with the loaded LLM and tokenizer
    """
    # def __init__(self,**kwargs):
    #     # self.model_uri =
    #     super().__init__(kwargs) ## this line is required
    model_uri = "EleutherAI/pythia-70m-deduped"
    def model_loader(self):
        return GPTNeoXForCausalLM.from_pretrained(self.model_uri)
    def tokenizer_loader(self):
        return AutoTokenizer.from_pretrained(self.model_uri)


@RegisterModelClass("opt")
class SmallLocalOpt(BaseOnsiteLLM):

    """
    This is a class for Facebook's OPT-350m LLM

    Attributes:
        model_uri (str): Hugging Face Endpoint for LLM
        tokenizer (AutoTokenizer): Tokenizer from Transformer's library
        model (LLM): The large language model

    Methods:
        model_loader: Loads the LLM into memory
        tokenizer_loader: Loads the tokenizer into memory
        generate: Generates a response from a given prompt with the loaded LLM and tokenizer
    """
    model_uri="facebook/opt-350m"
    def model_loader(self):
        return OPTForCausalLM.from_pretrained(self.model_uri)
    def tokenizer_loader(self):
        return AutoTokenizer.from_pretrained(self.model_uri)

@RegisterModelClass("bloom")
class SmallLocalBloom(BaseOnsiteLLM):

    """
    This is a class for BigScience's bloom-560 LLM

    Attributes:
        model_uri (str): Hugging Face Endpoint for LLM
        tokenizer (AutoTokenizer): Tokenizer from Transformer's library
        model (LLM): The large language model

    Methods:
        model_loader: Loads the LLM into memory
        tokenizer_loader: Loads the tokenizer into memory
        generate: Generates a response from a given prompt with the loaded LLM and tokenizer
    """
    model_uri="bigscience/bloom-560m"

    def model_loader(self):
        return BloomForCausalLM.from_pretrained(self.model_uri)
    def tokenizer_loader(self):
        return AutoTokenizer.from_pretrained(self.model_uri)

@RegisterModelClass("neo")
class SmallLocalNeo(BaseOnsiteLLM):
    """
    Attributes:
        model_uri (str): Hugging Face Endpoint for LLM
        tokenizer (AutoTokenizer): Tokenizer from Transformer's library
        model (LLM): The large language model

    Methods:
        model_loader: Loads the LLM into memory
        tokenizer_loader: Loads the tokenizer into memory
        generate: Generates a response from a given prompt with the loaded LLM and tokenizer
    """
    model_uri="EleutherAI/gpt-neo-1.3B"

    def model_loader(self):
        return GPTNeoForCausalLM.from_pretrained(self.model_uri)
    def tokenizer_loader(self):
        return AutoTokenizer.from_pretrained(self.model_uri)


# Add support for "Open-Orca/LlongOrca-7B-16k"
@RegisterModelClass("smallorca")
class SmallLocalOpenOrca(BaseOnsiteLLM):
    """
    This is a class for Openlm-Research's open_orca-3b LLM
    
    Attributes:
        model_uri (str): Hugging Face Endpoint for LLM
        tokenizer (AutoTokenizer): Tokenizer from Transformer's library
        model (LLM): The large language model
    
    Methods:
        model_loader: Loads the LLM into memory
        tokenizer_loader: Loads the tokenizer into memory
        generate: Generates a response from a given prompt with the loaded LLM and tokenizer
    """
    model_uri="Open-Orca/LlongOrca-7B-16k"
    
    def model_loader(self):
        return LlamaForCausalLM.from_pretrained(self.model_uri)
    def tokenizer_loader(self):
        return LlamaTokenizer.from_pretrained(self.model_uri)


# Add support for "Open-Orca/LlongOrca-13B-16k"
@RegisterModelClass("orca")
class LocalOpenOrca2(BaseOnsiteLLM):
    """
    This is a class for Openlm-Research's open_orca-3b LLM
        
    Attributes:
        model_uri (str): Hugging Face Endpoint for LLM
        tokenizer (AutoTokenizer): Tokenizer from Transformer's library
        model (LLM): The large language model
        
    Methods:
        model_loader: Loads the LLM into memory
        tokenizer_loader: Loads the tokenizer into memory
        generate: Generates a response from a given prompt with the loaded LLM and tokenizer
    """
    model_uri="Open-Orca/LlongOrca-13B-16k"
        
    def model_loader(self):
        return LlamaForCausalLM.from_pretrained(self.model_uri)
    def tokenizer_loader(self):
        return LlamaTokenizer.from_pretrained(self.model_uri)


# Add support for "Open-Orca/Mistral-7B-OpenOrca"
@RegisterModelClass("mistral")
class SmallLocalOpenMistral(BaseOnsiteLLM):
    """
    This is a class for OpenOrca's Mistral-7b LLM
                
    Attributes:
        model_uri (str): Hugging Face Endpoint for LLM
        tokenizer (AutoTokenizer): Tokenizer from Transformer's library
        model (LLM): The large language model
                
    Methods:
        model_loader: Loads the LLM into memory
        tokenizer_loader: Loads the tokenizer into memory
        generate: Generates a response from a given prompt with the loaded LLM and tokenizer
    """
    model_uri="Open-Orca/Mistral-7B-OpenOrca"
                
    def model_loader(self):
        return LlamaForCausalLM.from_pretrained(self.model_uri)
    def tokenizer_loader(self):
        return LlamaTokenizer.from_pretrained(self.model_uri) 


# Add support for "Open-Orca/OpenOrca-Platypus2-13B"
@RegisterModelClass("platypus")
class LocalOpenPlatypus(BaseOnsiteLLM):
    """
    This is a class for Open Orca's OpenOrca Platypus2-13b LLM
                
    Attributes:
        model_uri (str): Hugging Face Endpoint for LLM
        tokenizer (AutoTokenizer): Tokenizer from Transformer's library
        model (LLM): The large language model
                
    Methods:
        model_loader: Loads the LLM into memory
        tokenizer_loader: Loads the tokenizer into memory
        generate: Generates a response from a given prompt with the loaded LLM and tokenizer
    """
    model_uri="Open-Orca/OpenOrca-Platypus2-13B"
                
    def model_loader(self):
        return LlamaForCausalLM.from_pretrained(self.model_uri)
    def tokenizer_loader(self):
        return LlamaTokenizer.from_pretrained(self.model_uri)
    

@RegisterModelClass("llama")
class SmallLocalOpenLLama(BaseOnsiteLLM):
    """
    This is a class for Openlm-Research's open_llama-3b LLM

    Attributes:
        model_uri (str): Hugging Face Endpoint for LLM
        tokenizer (AutoTokenizer): Tokenizer from Transformer's library
        model (LLM): The large language model

    Methods:
        model_loader: Loads the LLM into memory
        tokenizer_loader: Loads the tokenizer into memory
        generate: Generates a response from a given prompt with the loaded LLM and tokenizer
    """
    model_uri="openlm-research/open_llama_3b_v2"

    def model_loader(self):
        return LlamaForCausalLM.from_pretrained(self.model_uri)
    def tokenizer_loader(self):
        return LlamaTokenizer.from_pretrained(self.model_uri)


@RegisterModelClass("llama2")
class SmallLocalLLama(BaseOnsiteLLM):

    """
    This is a class for Meta's llama-7b LLM

    Attributes:
        model_uri (str): Hugging Face Endpoint for LLM
        tokenizer (AutoTokenizer): Tokenizer from Transformer's library
        model (LLM): The large language model

    Methods:
        model_loader: Loads the LLM into memory
        tokenizer_loader: Loads the tokenizer into memory
        generate: Generates a response from a given prompt with the loaded LLM and tokenizer
    """
    model_uri="meta-llama/Llama-2-7b-hf"

    def model_loader(self):
        return LlamaForCausalLM.from_pretrained(self.model_uri)
    def tokenizer_loader(self):
        return LlamaTokenizer.from_pretrained(self.model_uri)


@RegisterModelClass("codellama-7b")
class CodeLlama7b(BaseOnsiteLLM):
        
    """
    This is a class for Meta's code-llama-7b LLM
    
    Attributes:
        model_uri (str): Hugging Face Endpoint for LLM
        tokenizer (AutoTokenizer): Tokenizer from Transformer's library
        model (LLM): The large language model
    
    Methods:
        model_loader: Loads the LLM into memory
        tokenizer_loader: Loads the tokenizer into memory
        generate: Generates a response from a given prompt with the loaded LLM and tokenizer
    """
    model_uri="codellama/CodeLlama-7b-hf"
    
    def model_loader(self):
        return LlamaForCausalLM.from_pretrained(self.model_uri)
    def tokenizer_loader(self):
        return CodeLlamaTokenizer.from_pretrained(self.model_uri)
    
@RegisterModelClass("codellama-13b")
class CodeLlama13b(BaseOnsiteLLM):
        
    """
    This is a class for Meta's code-llama-13b LLM
    
    Attributes:
        model_uri (str): Hugging Face Endpoint for LLM
        tokenizer (AutoTokenizer): Tokenizer from Transformer's library
        model (LLM): The large language model
    
    Methods:
        model_loader: Loads the LLM into memory
        tokenizer_loader: Loads the tokenizer into memory
        generate: Generates a response from a given prompt with the loaded LLM and tokenizer
    """
    model_uri="codellama/CodeLlama-13b-hf"
    
    def model_loader(self):
        return LlamaForCausalLM.from_pretrained(self.model_uri)
    def tokenizer_loader(self):
        return CodeLlamaTokenizer.from_pretrained(self.model_uri)   
    
@RegisterModelClass("codellama-34b")
class CodeLlama34b(BaseOnsiteLLM):
        
    """
    This is a class for Meta's code-llama-34b LLM
    
    Attributes:
        model_uri (str): Hugging Face Endpoint for LLM
        tokenizer (AutoTokenizer): Tokenizer from Transformer's library
        model (LLM): The large language model
    
    Methods:
        model_loader: Loads the LLM into memory
        tokenizer_loader: Loads the tokenizer into memory
        generate: Generates a response from a given prompt with the loaded LLM and tokenizer
    """
    model_uri="codellama/CodeLlama-34b-hf"
    
    def model_loader(self):
        return LlamaForCausalLM.from_pretrained(self.model_uri)
    def tokenizer_loader(self):
        return CodeLlamaTokenizer.from_pretrained(self.model_uri)   


@RegisterModelClass("flan")# our yummiest model based on similarity to food
class SmallLocalFlanT5(BaseOnsiteLLM):

    """
    This is a class for Google's flan-t5 LLM

    Attributes:
        model_uri (str): Hugging Face Endpoint for LLM
        tokenizer (AutoTokenizer): Tokenizer from Transformer's library
        model (LLM): The large language model

    Methods:
        model_loader: Loads the LLM into memory
        tokenizer_loader: Loads the tokenizer into memory
        generate: Generates a response from a given prompt with the loaded LLM and tokenizer
    """

    model_uri="google/flan-t5-small"
    def model_loader(self):
        return AutoModelForSeq2SeqLM.from_pretrained(self.model_uri)
    def tokenizer_loader(self):
        return AutoTokenizer.from_pretrained(self.model_uri)

@RegisterModelClass("bert")
class SmallLocalBERT(BaseOnsiteLLM):

    """
    This is a class for BERT: Pre-training of Deep Bidirectional Transformers for Language Understanding
    The base model needs finetuning in almost all cases.

    Attributes:
        model_uri (str): Hugging Face Endpoint for LLM
        tokenizer (AutoTokenizer): Tokenizer from Transformer's library
        model (LLM): The large language model

    Methods:
        model_loader: Loads the LLM into memory
        tokenizer_loader: Loads the tokenizer into memory
        generate: Generates a response from a given prompt with the loaded LLM and tokenizer
    """

    model_uri = "bert-base-cased"
    def model_loader(self):
        return AutoModelForMaskedLM.from_pretrained(self.model_uri)
    def tokenizer_loader(self):
        return AutoTokenizer.from_pretrained(self.model_uri)

@RegisterModelClass("gpt")
class GPT3:

    """
    This is a class for openAI's completion endpoint

    Methods:
        generate: Generates a response from a given prompt with OpenAI's completion endpoint
    """

    def generate(self,prompt, max_length=100,**kwargs): # both tokenizer and model take kwargs :(
        """
        This function uses openAI's API to generate a response from the prompt

        Parameters:
            prompt (str): Prompt to send to LLM
            max_length (int): Optional parameter limiting response length


        Returns:
            str: LLM Generated Response

        Example:
            >>> SmallLocalOpt.generate("How long does it take for an apple to grow?")
            It typically takes about 100-200 days...
        """

        ans = openai.Completion.create(prompt= prompt, model="text-davinci-003", **kwargs)
        return ans['choices'][0]['text']


    def finetune(self, dataset, optimizer, c_id, small_model_filename=None):
        old_model = optimizer.storage.get_model(c_id)
        training_file = create_jsonl_file(dataset)
        upload_response = openai.File.create(file=training_file, purpose="fine-tune", model="gpt-3.5-turbo-0613")
        training_file.close()
        fine_tuning_job = openai.FineTune.create(training_file= upload_response.id)

        print(f"Fine-tuning job created: {fine_tuning_job}", flush=True, file=sys.stderr)
        global job_id # global state isn't great, but thats interrupt handlers
        job_id = fine_tuning_job["id"]
        while True:
            fine_tuning_status = openai.FineTune.retrieve(id=job_id)
            status = fine_tuning_status["status"]
            print(f"Fine-tuning job status: {status}", file=sys.stderr)
            if status in ["succeeded", "completed", "failed"]:
                break
            time.sleep(30)
        job_id = None #
        new_model_id = fine_tuning_status.fine_tuned_model

        print("New_model_id: ", new_model_id, flush=True, file=sys.stderr)

        optimizer.storage.set_model(c_id, new_model_id)
        optimizer.storage.set_training_in_progress(c_id, False)
        if old_model is not None:
            openai.Model.delete(old_model)


@RegisterModelClass("gpt4")
class GPT4:
    """
    This is a class for openAI's gpt-4 LLM

    Methods:
        generate: Generates a response from a given prompt through OpenAI's endpoint
    """

    def generate(self, prompt, max_length=100, **kwargs):
        """
        This function uses openAI's API to generate a response from the prompt using the GPT-4 model

        Parameters:
            prompt (str): Prompt to send to LLM
            max_length (int): Optional parameter limiting response length


        Returns:
            str: LLM Generated Response

        Example:
            >>> GPT4.generate("How long does it take for an apple to grow?")
            It typically takes about 100-200 days...
        """

        cur_prompt = [{'role': "system", 'content': prompt}]
        ans = openai.ChatCompletion.create(
            messages=cur_prompt,
            model="gpt-4",
            **kwargs)

        return ans['choices'][0]['message']['content']

    def finetune(self, dataset, optimizer, c_id, small_model_filename=None):
        print("fine tuning isn't supported by OpenAI on this model", file=sys.stderr)
        raise Exception("fine tuning isn't supported by OpenAI on this model")


@RegisterModelClass("chat_gpt")
class ChatGPT:
    """
    This is a class for openAI's gpt-3.5-turbo LLM

    Methods:
        generate: Generates a response from a given prompt through OpenAI's endpoint
    """

    def generate(self,prompt, max_length=100,**kwargs): # both tokenizer and model take kwargs :(
        """
        This function uses openAI's API to generate a response from the prompt

        Parameters:
            prompt (str): Prompt to send to LLM
            max_length (int): Optional parameter limiting response length


        Returns:
            str: LLM Generated Response

        Example:
            >>> SmallLocalOpt.generate("How long does it take for an apple to grow?")
            It typically takes about 100-200 days...
        """
        cur_prompt = [{'role': "system", 'content' : prompt}]
        ans = openai.ChatCompletion.create(
            messages=cur_prompt,
            model="gpt-3.5-turbo-0301",
            **kwargs)
        return ans['choices'][0]['message']['content']

    def finetune(self, dataset, optimizer, c_id, small_model_filename=None):
        print("fine tuning isn't supported by OpenAI on this model", file=sys.stderr)
        raise Exception("fine tuning isn't supported by OpenAI on this model")
        # old_model = optimizer.storage.get_model(c_id)
        # training_file = create_jsonl_file(dataset)
        # upload_response = openai.File.create(file=training_file, purpose="fine-tune")
        # training_file.close()
        # fine_tuning_job = openai.FineTune.create(training_file= upload_response.id)

        # print(f"Fine-tuning job created: {fine_tuning_job}", flush=True)
        # global job_id # global state isn't great, but thats interrupt handlers
        # job_id = fine_tuning_job["id"]
        # while True:
        #     fine_tuning_status = openai.FineTune.retrieve(id=job_id)
        #     status = fine_tuning_status["status"]
        #     print(f"Fine-tuning job status: {status}")
        #     if status in ["succeeded", "completed", "failed"]:
        #         break
        #     time.sleep(30)
        # job_id = None #
        # new_model_id = fine_tuning_status.fine_tuned_model

        # print("New_model_id: ", new_model_id, flush=True)

        # optimizer.storage.set_model(c_id, new_model_id)
        # optimizer.storage.set_training_in_progress(c_id, False)
        # if old_model is not None:
        #     openai.Model.delete(old_model)


class BaseCtransformersLLM(BaseOnsiteLLM):
    """
    Base Class for running Ctransformers/GGML models

    Attributes:
        model_uri (str): Ctransformers uri for LLM
        model_kwargs (dict): Keyword arguments for loading the LLM
    
    Methods:
        model_loader: Loads specified model from Ctransformers
        generate: Generates a response from given prompt

    """

    def __init__(self, with_GPU=False, **model_kwargs):
        self.__model_uri = None
        self.__model_file = None
        if with_GPU:
            self.model = self.gpu_model_loader(**model_kwargs)
        else:
            self.model = self.model_loader(**model_kwargs)

    @property
    def model_file(self):
        return self.__model_file

    @model_file.setter
    def model_file(self,val):
        self.__model_file=val 

    @property
    def model_uri(self):
        return self.__model_uri

    @model_uri.setter
    def model_uri(self,val):
        self.__model_uri=val 
        
    def load_finetune(self, model_filename):
            raise Exception("Finetuning not supported for Ctransformers/GGML.")

    def _get_model_layers(self):
        pass

    def _get_model_size(self):
        pass    
    
    def model_loader(self, **kwargs):
        if 'model_file' in kwargs:
            del kwargs['model_file']

        if self.model_file is not None:
            return AutoModelForCausalLM.from_pretrained(self.__model_uri, model_file=self.__model_file, **kwargs)
        else:
            return AutoModelForCausalLM.from_pretrained(self.__model_uri, **kwargs)
        
    def gpu_model_loader(self, vram=0, **kwargs):
        if 'model_file' in kwargs:
            del kwargs['model_file']

        if vram > 0:
            model_size = self._get_model_size()
            model_layers = self._get_model_layers()

            size_per_layer = model_size // model_layers
            offload_layers = vram // size_per_layer
            if offload_layers > model_layers:
                offload_layers = model_layers

            if self.model_file is not None:
                return AutoModelForCausalLM.from_pretrained(self.__model_uri, model_file=self.__model_file, gpu_layers=offload_layers, **kwargs)
            else:
                return AutoModelForCausalLM.from_pretrained(self.__model_uri, gpu_layers=offload_layers, **kwargs)
        else: 
            raise ValueError("Expected VRAM to be greater than 0.")
        
    def generate(self, prompt, *generate_kwargs):  
        input_ids = self.model.tokenize(prompt)
        response = self.model.generate(input_ids, *generate_kwargs)
        return self.model.detokenize(response)
    
    def finetune(self, data, optimizer, c_id, model_filename=None):
        raise Exception("Finetuning not supported for Ctransformers/GGML.")


@RegisterModelClass("quantized-llama2-7b-base")
class Base_Llama2_7b_Q4(BaseCtransformersLLM):
    """
    Class for running quantized Llama 2 7b base model instance

    Properties:
        model_uri: Ctransformers uri for LLM
        model_file: gguf or bin file for repos with multiple files
    """

    model_uri="TheBloke/Llama-2-7B-GGML"
    model_file="llama-2-7b.ggmlv3.q4_K_M.bin"

@RegisterModelClass("quantized-llama2-13b-base")

class Base_Llama2_13b_Q4(BaseCtransformersLLM):
    """
    Class for running quantized Llama 2 13b base model instance

    Properties:
        model_uri: Ctransformers uri for LLM
        model_file: gguf or bin file for repos with multiple files
    """

    model_uri="TheBloke/Llama-2-13B-GGML"
    model_file="llama-2-13b.ggmlv3.q4_K_M.bin"


@RegisterModelClass("llama2-7b-chat-Q4")
class Chat_Llama2_7b_Q4(BaseCtransformersLLM):
    """
    Class for running Llama2-7b-Chat model instance

    Properties:
        model_uri: Ctransformers uri for LLM
        model_file: gguf or bin file for repos with multiple files
    """

    model_uri="TheBloke/Llama-2-7B-Chat-GGML"
    model_file="llama-2-7b-chat.Q4_K_M.gguf"

@RegisterModelClass("llama2-7b-chat-Q6")
class Chat_Llama2_7b_Q6(BaseCtransformersLLM):
    """
    Class for running Llama2-7b-Chat model instance

    Properties:
        model_uri: Ctransformers uri for LLM
        model_file: gguf or bin file for repos with multiple files
    """

    model_uri="TheBloke/Llama-2-7B-Chat-GGML"
    model_file="llama-2-7b-chat.Q6_K.gguf"
 

@RegisterModelClass("llama2-13b-chat-Q4")
class Chat_Llama2_13b_Q4(BaseCtransformersLLM):
    """
    Class for running Llama2-13b-Chat model instance

    Properties:
        model_uri: Ctransformers uri for LLM
        model_file: gguf or bin file for repos with multiple files
    """

    model_uri="TheBloke/Llama-2-13B-Chat-GGML"
    model_file="llama-2-13b-chat.Q4_K_M.gguf"


@RegisterModelClass("llama2-13b-chat-Q6")
class Chat_Llama2_13b_Q6(BaseCtransformersLLM):
    """
    Class for running Llama2-13b-Chat model instance

    Properties:
        model_uri: Ctransformers uri for LLM
        model_file: gguf or bin file for repos with multiple files
    """

    model_uri="TheBloke/Llama-2-13B-Chat-GGML"
    model_file="llama-2-13b-chat.Q6_K.gguf"


@RegisterModelClass("llama2-7b-32k-Q4")
class Chat_Llama2_13b_Q6(BaseCtransformersLLM):
    """
    Class for running Llama2-7b-32k-Instruct model instance

    Properties:
        model_uri: Ctransformers uri for LLM
        model_file: gguf or bin file for repos with multiple files
    """

    model_uri="TheBloke/Llama-2-7B-32K-Instruct-GGML"
    model_file="llama-2-7b-32k-instruct.ggmlv3.q4_1.bin"



"""
API Key Definitions and Regex Pattern for Pure Interpolations.

This section defines and retrieves the API keys required for various services from the environment variables.
It also includes a regular expression pattern used to identify pure interpolations.

API Key Definitions:
- OPENAI_DEFAULT_KEY: The API key for OPENAI.
- GOOGLE_MAPS_KEY: The API key for Google Maps.
- SERPAPI_KEY: The API key for SERPAPI (Anarchy AI)
- WOLFRAM_KEY: The API key for Wolfram.


Note: The corresponding environment variables should be set before running
any of the agents. Set the environment variables without quotes.
E.g.: WOLFRAM_KEY=an-encypted-key
"""
import os
import openai



###
# Use this regex to find if a variable is a pure interpolation.
# For @example, { "name": "{my_name}" } is pure, since there's nothing else besides the interpolation
# For @example, { "Ticker": "TickerID={ticker_id}|historical=1" } is NOT pure, since there is extra text
DICT_KEY_REGEX_TO_FIND_PURE_INTERPOLATIONS = "^\{[a-zA-Z0-9\.\-_]+\}$"


def set_api_key(key, key_type="OPENAI_API_KEY"):
    """
    Set the API key in the environment variable.

    Parameters:
    - key:  The API key to set. Defaults to OPENAI_DEFAULT_KEY.
            There are four options:
                - OPENAI_DEFAULT_KEY
                - GOOGLE_MAPS_KEY
                - GOOGLE_KEY
                - GOOGLE_CX
    - key_type: The type of API key. Defaults to "OPENAI_API_KEY".
                There are four options:
                - OPENAI_API_KEY
                - GOOGLE_MAPS_KEY
                - GOOGLE_KEY
                - GOOGLE_CX
    """
    if not os.getenv(key_type):
        os.environ[key_type]=key

        if key_type == "OPENAI_API_KEY":
            openai.api_key=str(key)



"""
This Python script runs the REBEL agent, which takes into account of multiple questions within a
single user input.
The user is prompted to enter a question, to which the AI responds.
If multiple questions are present, the script splits these into different subquestions and retrieves
answers from the respective APIs.
The script also takes into account of previous prompts as history, in case the user may
enter related questions later.
The user will also see messages regarding which API for information was chosen and the
price of the API call.
"""

import os
import sys

# Get the current file's directory to grab the python files with common functionality in the utils/ folder
# current_dir = os.path.dirname(os.path.abspath(__file__))
# grandparent_dir = os.path.dirname(os.path.dirname(current_dir))
# utils_dir = os.path.join(grandparent_dir, "utils/")
# sys.path.append(utils_dir)

from llm_vm.utils.keys import *
from llm_vm.utils.labels import *
from llm_vm.utils.print_types import *
from llm_vm.utils.tools import *
from urllib.parse import urlencode
import urllib.parse as urlparse
import json
import random

# UNCOMMENT THESE IMPORTS SO THAT SUBQUESTION PORTION CAN WORK (NLP CURRENTLY NOT RECOGNIZED)
import spacy

nlp = None #spacy.load("en_core_web_md")

from math import sqrt, pow, exp

try:
    from .bothandler import (
        question_split,
        tool_picker,
        memory_check,
        replace_variables_for_values,
    )

except:
    from bothandler import (
        question_split,
        tool_picker,
        memory_check,
        replace_variables_for_values,
    )

try:
    from .utils import *
except:
    from utils import *


def prepPrintPromptContext(p):
    return "--> " + p.replace("\n", "\n--> ") if len(p) > 0 else ""


def print_op(*kargs, **kwargs):
    print(*kargs, **kwargs, flush=True, file=sys.stderr)


random_fixed_seed = random.Random(4)

QUALITY = 0.2


def buildExampleTools(tools=GENERIC_TOOLS):
    new_tools = tools

    # wolfram
    new_tools[0]["examples"] = [
        (
            [
                (
                    "What's the most popular spot to vacation if you are in Germany?",
                    "Mallorca",
                ),
                ("What's the date", "July 7, 2009"),
            ],
            "How crowded is it there now?",
            '{"q": "How many tourists visit Mallorca each July?"}',
        ),
        (
            [
                ("What is the circumference of a basketball?", "750mm"),
            ],
            "What is the volume of a basketball?",
            '{"q": "What is the volume of a ball with circumference 750mm?"}',
        ),
        (
            [
                ("What's the fastest a ford explorer can drive?", "160mph"),
            ],
            "What is that multiplied by seventy?",
            '{"q": "160 x 70"}',
        ),
        (
            [],
            "Is 5073 raised to the 3rd power divisible by 73?",
            '{"q": "Is 5073^3 divisible by 73?"}',
        ),
    ]

    # geopy
    new_tools[1]["examples"] = [
        (
            [
                (
                    "I feel like taking a drive today to zurich can you help?",
                    "Yes!  Where are you now?",
                ),
            ],
            "I'm in Paris.",
            '{"origins": "Paris", "destinations": "Zurich"}',
        ),
        (
            [],
            "How long would it take to get between South Africa and Kenya.",
            '{"origins": "South Africa", "destinations": "Kenya"}',
        ),
        (
            [
                ("Where do elephant seals mate?", "San Luis Obispo"),
                ("Thats really cool!", "Damn right."),
                ("Where do they migrate each year?", "Alaska"),
            ],
            "How many miles would they travel while doing that?",
            '{"origins": "San Luis Obispo", "destinations": "Alaska"}',
        ),
    ]

    # weather
    new_tools[2]["examples"] = [
        (
            [
                ("Are you allergic to seafood?", "I'm just an AI. I have no body."),
                ("What city is Obama in?", "NYC"),
                ("What is the latitude of NYC?", "40.7128° N"),
                (
                    "What is the longitude of NYC?",
                    "The longitude is 73° 56' 6.8712'' W",
                ),
            ],
            "What is the weather in NYC?",
            '{"longitude": 73.56, "latitude": 40.41728}',
        ),
        (
            [
                ("What is the longitude of Milan?", "9.1900° E"),
                ("What is the latitude of Milan?", "45.4642° N"),
            ],
            "What is the weather in Milan?",
            '{"longitude": 45.4642, "latitude": 9.1900}',
        ),
    ]
    return new_tools


def squared_sum(x):
    """return 3 rounded square rooted value"""

    return round(sqrt(sum([a * a for a in x])), 3)


def cos_similarity(x, y):
    """return cosine similarity between two lists"""

    numerator = sum(a * b for a, b in zip(x, y))
    denominator = squared_sum(x) * squared_sum(y)
    return round(numerator / float(denominator), 3)


class Agent:
    def __init__(self, openai_key, tools, bot_str="", verbose=4):
        self.verbose = verbose
        self.price = 0
        self.tools = []
        self.set_tools(buildExampleTools()+tools)
        if bot_str == "":
            self.bot_str = bot_str
        else:
            self.bot_str = "<GLOBAL>" + bot_str + "<GLOBAL>"
        self.nlp=spacy.load("en_core_web_md")
        # set all the API resource keys to make calls
        set_api_key(openai_key, "OPENAI_API_KEY")

    def makeToolDesc(self, tool_id):
        """
        Creates the tool description to contain the relevant tags according to the dynamic params

        Parameters
        ----------
        tool_id
            the tool's enum value as specified in the DefaultTools class

        Returns
        ----------
        String
            a formatted string with tags containing the tool_id, description and params
        """

        tool = self.tools[tool_id]
        params = (
            "{"
            + ", ".join(
                ['"' + l + '": ' + v for l, v in tool["dynamic_params"].items()]
            )
            + "}"
            if tool["dynamic_params"] != ""
            else "{}"
        )
        return f"""<TOOL>
    <{TOOL_ID}>{str(tool_id)}</{TOOL_ID}>
    <{DESCRIPTION}>{tool['description']}</{DESCRIPTION}>
    <{PARAMS}>{params}</{PARAMS}>
    </TOOL>"""

    def set_tools(self, tools):
        """
        Adds all the available tools when the class is initialized.

        Parameters
        ----------
        tools
            a list of available tools. Each tool is defined within a Dict.

        Returns
        ----------
        None
        """

        for tool in tools:
            if not "args" in tool:
                tool["args"] = {}
            if not "method" in tool:
                tool["method"] = "GET"
            if not "examples" in tool:
                tool["examples"] = []
            if not "dynamic_params" in tool:
                tool["dynamic_params"] = {}
            self.tools += [tool]

    def use_tool(self, tool, gpt_suggested_input, question, memory, facts, query=""):
        """
        Calls the appropriate API based on the tool that's in use.

        Parameters
        ----------
        tool
            an integer refencing the selected tool
        gpt_suggested_input
            the input params for the selected tool as specified by the gpt response
        question
            the user's input question
        memory
            a list of tuples containing the conversation history
        facts
            a list containing factual answers generated for each sub-question
        query
            the value of the ai_response_prompt key for the selected tool. If ai_response_prompt key is not available this will have same value as question parameter

        Returns
        ----------
        String
            Response text after calling API tool and passing result into ChatGPT prompt
        """

        return tool_api_call(
            self, tool, gpt_suggested_input, question, memory, facts, query=query
        )

    def run(self, question, memory):
        """
        Runs the Agent on the given inputs

        Parameters
        ----------
        question
            the user's input question
        memory
            a list of tuples containing the conversation history

        Returns
        ----------
        Tuple
            a tuple containing the Agent's answer and a list of the conversation history
        """

        self.price = 0

        thought, facts = self.promptf(
            question,
            memory,
            [],
            0
        )

        if self.verbose > -1:
            print_big("GPT-3.5 Price = ~{:.1f} cents".format(self.price * 100))

        # return (answer, memory + [(question, answer)], calls, debug_return, has_friendly_tags)
        return (thought, memory + [(question, thought)])

    def makeInteraction(self, p, a, Q="HUMAN", A="AI", INTERACTION=INTERACTION):
        """
        Formats the tool description to contain the relevant tags according to the dynamic params

        Parameters
        ----------
        p
            the question being asked
        a
            the gpt response
        Q
            the entity asking the question. Could be HUMAN or AI.
        A
            the entity answering the question. Could be HUMAN or AI.
        INTERACTION
            the type of interaction. Could be HUMAN-AI or AI-AI.

        Returns
        ----------
        String
            a formatted string with tags containing the type of interaction, the question and the gpt response
        """

        return f"<>{INTERACTION}:<{Q}>{p}</{Q}>\n<{A}>" + (
            f"{a}</{A}>\n</>" if a is not None else ""
        )

    def make_sub(
        self,
        tools,
        memory,
        facts,
        question,
        subq,
        answer_label,
        toolEx,
        tool_to_use=None,
        bot_str="",
        quality="best",
        max_tokens=20,
    ):
        """
        Formats the tool description to contain the relevant tags according to the dynamic params

        Parameters
        ----------
        tools
            a list of available tools. Each tool is defined within a Dict.
        memory
            a list of tuples containing the conversation history
        facts
            a list containing factual answers generated for each sub-question
        question
            the user's input question
        subq
            a lambda function that returns a question that's used to prompt gpt for the suggested input for the selected tool
        answer_label
            the format of the response
        toolEx
            a lambda function that returns an example for the selected tool
        tool_to_use
            an integer referencing the selected tool
        bot_str
            a string to be appended to the gpt prompt
        quality
            could be either "okay"-text-curie-001 or "best"-text-davinci-003
        max_tokens
            maximum number of tokens to be generated

        Returns
        ----------
        String
            a gpt text response containing the suggested input for the selected tool
        """
        tool_context = "".join([self.makeToolDesc(t) for t, _ in tools])
        mem = "".join(
            [
                self.makeInteraction(p, a, "P", "AI", INTERACTION="Human-AI")
                for p, a in memory
            ]
        ) + "".join(
            [
                self.makeInteraction(p, a, "P", "AI", INTERACTION="AI-AI")
                for p, a in facts
            ]
        )

        def makeQuestion(memory, question, tool=None):
            return (
                f"\n\n<{EXAMPLE}>\n"
                + f"<{QUESTION}>{question}</{QUESTION}>\n"
                + f"<{THOUGHT}><{PROMPT}>{subq(tool)}</{PROMPT}>\n<{RESPONSE} ty={answer_label}>\n"
            )

        examples = []
        for tool_id, tool in tools:
            for tool_example in tool["examples"]:
                examples += [
                    makeQuestion(tool_example[0], tool_example[1], tool_id)
                    + toolEx(tool_id, tool_example)
                    + f"</{RESPONSE}></{THOUGHT}></{EXAMPLE}>"
                ]

        random_fixed_seed.shuffle(examples)

        cur_question = f"<{CONVERSATION}>{mem}</{CONVERSATION}>" + makeQuestion(
            memory, question, tool=tool_to_use
        )
        prompt = MSG("system", "You are a good and helpful assistant.")
        prompt += MSG(
            "user",
            "<TOOLS>" + tool_context + "</TOOLS>" + "".join(examples) + cur_question,
        )
        return call_ChatGPT(
            self, prompt, stop=f"</{RESPONSE}>", max_tokens=max_tokens
        ).strip()

    def promptf(self, question, memory, facts, level, split_allowed=True, max_level=3):
        """
        Formats the gpt prompt to include conversation history, facts and logs responses to the console

        Parameters
        ----------
        question
            the user's input question
        memory
            a list of tuples containing the conversation history
        facts
            a list containing factual answers generated for each sub-question
        level
            param indicating the current recursive level

        split_allowed
            a boolean to allow the question to be split into sub-questions if set to True

        max_level
            param that indicates the maximum recursive level we want to allow.

        Returns
        ----------
        Tuple
            a tuple containing the gpt response and a list with the conversation history
        """

        mem = "".join(
            [
                self.makeInteraction(p, a, "P", "AI", INTERACTION="Human-AI")
                for p, a in memory
            ]
        ) + "".join(
            [
                self.makeInteraction(p, a, "P", "AI", INTERACTION="AI-AI")
                for p, a in facts
            ]
        )

        if split_allowed:
            subq = question_split(question, self.tools, mem)

            subq_final = []

            if len(subq[1]) == 1:
                split_allowed = False

            else:
                for i in subq[1]:
                    if cos_similarity(self.nlp(question).vector, self.nlp(i).vector) < 0.98:
                        subq_final.append(i)
                    else:
                        split_allowed = False

            if level==max_level:
                split_allowed = False

            self.price += subq[0]
            new_facts = []

            for i in range(len(subq_final)):
                _, new_facts = self.promptf(
                    subq_final[i],
                    memory,
                    facts,
                    level+1,
                    split_allowed=split_allowed,
                )
                facts = facts + new_facts
                mem = "".join(
                    [
                        self.makeInteraction(p, a, "P", "AI", INTERACTION="Human-AI")
                        for p, a in memory
                    ]
                ) + "".join(
                    [
                        self.makeInteraction(p, a, "P", "AI", INTERACTION="AI-AI")
                        for p, a in facts
                    ]
                )

        answer_in_memory = memory_check(mem, question)
        self.price += answer_in_memory[0]
        answer_in_memory = answer_in_memory[1]
        if answer_in_memory:
            prompt = MSG("system", "You are a good and helpful bot" + self.bot_str)
            prompt += MSG(
                "user",
                mem + "\nQ:" + question + "\nANSWER Q, DO NOT MAKE UP INFORMATION.",
            )
            a = call_ChatGPT(self, prompt, stop="</AI>", max_tokens=256).strip()
            print_big("NO DATA TOOLS USED, ANSWERED FROM MEMORY")
            return (a.replace("\n", ""), [(question, a)])

        tool_to_use = tool_picker(self.tools, question, 3)


        print_big(
            "".join(
                [
                    f'{"✓" if self.tools.index(tool) == tool_to_use else " "} {self.tools.index(tool)}.- {tool["description"]}\n'
                    for tool in self.tools[3:0]
                ]
            )
            + f"\n> Question: {question}\n> Raw answer: '{tool_to_use}'\n> Tool ID: {tool_to_use}",
            "LIST OF DATA TOOLS",
        )

        self.price += tool_to_use[0]
        try:
            tool_to_use = int(tool_to_use[1])
        except:
            tool_to_use = len(self.tools)
        if tool_to_use >= len(self.tools):
            prompt = MSG("system", "You are a good and helpful bot" + self.bot_str)
            prompt += MSG(
                "user",
                mem
                + "\nQ:"
                + question
                + "\nUsing this information, what is the answer to Q?",
            )
            a = call_ChatGPT(self, prompt, stop="</AI>", max_tokens=256).strip()

            return (a.replace("\n", ""), [(question, a)])
        tool_input = self.make_sub(
            list(enumerate(self.tools)),
            memory,
            facts,
            question,
            lambda t: "What should the input for tool " + str(t) + " be to answer Q?",
            "JSON",
            lambda t, ex: ex[2],
            tool_to_use=tool_to_use,
            quality="best" if self.price < QUALITY else "okay",
            max_tokens=200,
        )

        query = question
        if "ai_response_prompt" in self.tools[tool_to_use].keys():
            query = self.tools[tool_to_use]["ai_response_prompt"]
        try:
            answer = self.use_tool(
                self.tools[tool_to_use], tool_input, question, memory, facts, query=query
            )
        except:
            prompt = MSG("system", "You are a good and helpful bot" + self.bot_str)
            prompt += MSG(
                "user",
                mem + "\nQ:" + question + "\nANSWER Q, DO NOT MAKE UP INFORMATION.",
            )
            answer = call_ChatGPT(self, prompt, stop="</AI>", max_tokens=256).strip()


        return (answer, [(question, answer)])

def rebel_main():
    tools = buildExampleTools()


    label = Agent(os.getenv("LLM_VM_OPENAI_API_KEY"), tools, verbose=1)
    conversation_history = []
    last = ""
    while True:
        inp = input(last + "Human: ")
        return_value = label.run(inp, conversation_history)
        conversation_history = return_value[1]
        last = "AI: " + str(return_value[0]) + "\n"

if __name__ == "__main__":
    rebel_main()



"""
This file contains various machine learning classes and variables
for training and API calls to OpenAI.
"""

from typing import Tuple, List, Dict, TypedDict, Union, Any, Optional
from enum import Enum

class OpenAIModel(Enum):
    CURIE = "curie"
    CURIE_TEXT = "text-curie-001"
    FAST_DAVINCI = "code-cushman-002"
    DAVINCI = "davinci"
    DAVINCI_TEXT = "text-davinci-003"

class LLMCallType(Enum):
    OPENAI_COMPLETION = "openai-completion"
    OPENAI_CHAT = "openai-chat"
    # ... other models

class DecisionStep(Enum):
    SPLIT = "split_questions"
    INPUT = "guess_api_input"
    FROM_MEMORY = "use_memory"
    TOOL_PICKER = "tool_picker"

class DefaultTools(Enum):
    I_DONT_KNOW = -1
    ANSWER_FROM_MEMORY = 0

    WOLFRAM = 1
    DIRECTIONS = 2
    WEATHER = 3

    ## Always set this to the latest ID. This is an integer, not an enum!
    __LAST_ID__: int = 4

class DefaultTrainingTools(Enum):
    TRAIN_SEND_EMAIL = 4
    TRAIN_CREATE_DOCUMENT = 5
    TRAIN_SHARE_DOCUMENT = 6
    TRAIN_BOOK_TABLE = 7
    TRAIN_FIND_RESTAURANT = 8
    TRAIN_NEWS = 9
    TRAIN_ORDER_FOOD = 10
    TRAIN_BOOK_APPOINTMENT = 11
    TRAIN_AVAILABLE_APPOINTMENT = 12
    TRAIN_FIND_FLIGHT = 13

TupleList = List[Tuple[str, str]]

class ToolTypeArgs(TypedDict):
    url: str
    params: Optional[dict]
    json: Optional[dict]
    jsonParams: Optional[dict]
    auth: Optional[dict]
    cert: Optional[str]
    data: Optional[dict]

class SingleTool(TypedDict):
    description: str
    dynamic_params: dict
    id: int
    args: ToolTypeArgs
    ai_response_prompt: Union[str, None]

ToolList = List[SingleTool]
# ToolList = NewType("ToolList", List[SingleTool])
EnumeratedToolList = List[Tuple[int,SingleTool]]

ALL_TOOLS_MAP: Dict[str, Union[DefaultTools, DefaultTrainingTools]] = {
    "dont_know": DefaultTools.I_DONT_KNOW,
    "memory": DefaultTools.ANSWER_FROM_MEMORY,
    "weather": DefaultTools.WEATHER,
    "wolfram": DefaultTools.WOLFRAM,
    "directions": DefaultTools.DIRECTIONS,
    "send_email": DefaultTrainingTools.TRAIN_SEND_EMAIL,
    "create_doc": DefaultTrainingTools.TRAIN_CREATE_DOCUMENT,
    "share_doc": DefaultTrainingTools.TRAIN_SHARE_DOCUMENT,
    "book_table": DefaultTrainingTools.TRAIN_BOOK_TABLE,
    "find_restaurant": DefaultTrainingTools.TRAIN_FIND_RESTAURANT,
    "news": DefaultTrainingTools.TRAIN_NEWS,
    "order_food": DefaultTrainingTools.TRAIN_ORDER_FOOD,
    "book_appointment": DefaultTrainingTools.TRAIN_BOOK_APPOINTMENT,
    "check_appointment": DefaultTrainingTools.TRAIN_AVAILABLE_APPOINTMENT,
    "find_flight": DefaultTrainingTools.TRAIN_FIND_FLIGHT
}

class QuestionSplitModelJSONData(TypedDict):
    mem: List[List[str]]
    question: str
    answer: List[str]

class QuestionSplitModelData(TypedDict):
    mem: Union[TupleList, None]
    question: str
    answer: Union[List[str], None]
    tools: Union[ToolList, None]

class AnswerFromMemoryModelData(TypedDict):
    mem: Union[TupleList, None]
    question: str
    thought: Union[str, None]
    answer: Union[str, None]

class QuestionSplitInputModel(TypedDict):
    model_name: str
    openai_model: OpenAIModel
    data: List[QuestionSplitModelData]

class PromptModelEntry(TypedDict):
    prompt: str
    completion: str

class LLMCallParams(TypedDict):
    llm: LLMCallType
    model: Tuple[str, Union[str, None]]
    prompt: str
    temperature: float
    max_tokens: int
    stop: str

LLMCallReturnType = Tuple[str, float]

class LLMTrainingResult(TypedDict):
    elapsed_time_s: int
    model_name: str
    model_files: List[str]

class LLMTrainingResultMeta(TypedDict):
    OPENAI_KEY: str
    DATE: str
    TIMESTAMP: int

LLModels = Union[Dict[str, LLMTrainingResult], LLMTrainingResultMeta]


class ToolInputModelJSONData(TypedDict):
    mem: Optional[List[str]]
    description: str
    question: str
    answer: dict
    params: dict
class ToolInputModelData(TypedDict):
    mem: Optional[TupleList]
    description: str
    question: str
    answer: dict
    params: dict

class ToolInputModel(TypedDict):
    openai_model: OpenAIModel
    data: List[ToolInputModelData]

class ToolpickerInputModelJSONData(TypedDict):
    mem: Optional[List[str]]
    thought: Optional[str]
    question: str
    answer: str
class ToolpickerInputModelData(TypedDict):
    mem: Optional[TupleList]
    thought: Optional[str]
    question: str
    answer: int
    tools: Optional[ToolList]

class ToolpickerInputModel(TypedDict):
    openai_model: OpenAIModel
    data: List[ToolpickerInputModelData]

class AnswerInMemoryModelData(TypedDict):
    mem: Optional[TupleList]
    facts: Optional[TupleList]
    question: str
    answer: Optional[bool]

class AnswerInMemoryModel(TypedDict):
    openai_model: OpenAIModel
    data: List[AnswerInMemoryModelData]

class DebugCalls(TypedDict):
    response: str
    gpt_suggested_params: Any
    url: str
    raw_api_return: str

DebugCallList = List[DebugCalls]



import json
import sys
import openai
import time
import os
from llm_vm.utils.typings_llm import *
from llm_vm.agents.FLAT.models.helpers.check_model_status import check_model_status

__START_LABEL    = " 🎛️  Creating model..."
__PROGRESS_LABEL = " ⏳ Processing"
__FINISHED_LABEL = "\n 🎉 Finished!" + \
                   "\n 🔗"

def upload_model (
    data: PromptModelEntry,
    openai_model: str,
    file_name: str,
    is_test: bool = False
) -> LLMTrainingResult:
    start_of_execution = time.time()

    def progress_label(label: str) -> str:
        ellapsed_s = int(time.time() - start_of_execution)
        return f"{label} '{file_name}'\t ({'%dm %02ds' % (ellapsed_s / 60, ellapsed_s % 60)}) "

    current_dir = os.path.dirname(os.path.abspath(__file__))
    file_path = os.path.join(current_dir, f'./../data/{file_name}.jsonl')

    with open(file_path, "w") as output_file:
        output_file.truncate(0)
        for entry in data:
            json.dump(entry, output_file)
            output_file.write("\n")

    if is_test:
        return {
            "model_name": file_name + "_EXXX",
            "model_files": [],
            "elapsed_time_s": int(time.time() - start_of_execution)
        }

    upload_response = openai.File.create(
        file=open(file_path, "rb"),
        purpose='fine-tune'
    )

    file_id = upload_response.id

    fine_tune_response = openai.FineTune.create(training_file=file_id, model=openai_model)
    job_id = fine_tune_response["id"]
    print(__START_LABEL, f'''job_id: {job_id}, model: {file_name}''', flush=True, file=sys.stderr)

    finetuned_model = {"status": None}
    while finetuned_model["status"] not in ["succeeded", "failed"]:
        time.sleep(15)
        finetuned_model = check_model_status(job_id, progress_label(__PROGRESS_LABEL))

    print(progress_label(__FINISHED_LABEL), finetuned_model["fine_tuned_model"], flush=True, file=sys.stderr)

    related_files = (
        [file["id"] for file in finetuned_model["result_files"]] + \
        [file["id"] for file in finetuned_model["training_files"]]
    )

    return {
        "model_name": finetuned_model["fine_tuned_model"],
        "model_files": related_files,
        "elapsed_time_s": int(time.time() - start_of_execution)
    }



import json
import os
from llm_vm.utils.typings_llm import *

def __question_splitter_data() -> QuestionSplitInputModel:
    current_dir = os.path.dirname(os.path.abspath(__file__))
    file_path = os.path.abspath(os.path.join(current_dir, f'../../raw_data/question_split_data.json'))

    json_data: List[QuestionSplitModelJSONData] = json.load(open(file_path, "r"))["data"]

    question_data: List[QuestionSplitModelData] = [
        {
            "mem": [(a[0], a[1]) for a in q["mem"]] if "mem" in q else [],
            "question": q["question"],
            "answer": q["answer"]
        }
        for q in json_data
    ]

    return {
        "openai_model": OpenAIModel.DAVINCI,
        "data": question_data
    }

question_splitter_data = __question_splitter_data();



from abc import ABC,abstractmethod
import pinecone
import weaviate
from dotenv import load_dotenv
import os
import hnswlib

load_dotenv("../../.env.example")

class VectorDB(ABC):

    @abstractmethod
    def create_index(self):
        pass

    @abstractmethod
    def list_indexes(self):
        pass

    @abstractmethod
    def describe_index(self, index_name):
        pass

    @abstractmethod
    def delete_index(self, index_name):
        pass

    @abstractmethod
    def upsert(self, **kwargs):
        pass

    @abstractmethod
    def query(self, **kwargs):
        pass


class PineconeDB(VectorDB):
    def __init__(self, api_key, pinecone_env):
        self.pinecone = pinecone
        self.pinecone.init(api_key=api_key, environment=pinecone_env)
        self.index = None
        print(self.pinecone.list_indexes())
    
    def create_index(self, **kwargs):
        if "name" not in kwargs:
            raise ValueError("Expected name as a keyword argument but found None")
        if "dimension" not in kwargs:
            kwargs["dimension"] = 1024
        if "metric" not in kwargs:
            kwargs["metric"] = "cosine"

        self.pinecone.create_index(**kwargs)
        self.index = self.pinecone.Index(kwargs["name"])

    def list_indexes(self):
        return self.pinecone.list_indexes()
    
    def describe_index(self, index_name):
        return self.pinecone.describe_index(index_name)
    
    def delete_index(self, index_name):
        self.pinecone.delete_index(index_name)
        print(f"${index_name} has been deleted")

    def upsert(self, **kwargs):
        if "vectors" not in kwargs:
            raise ValueError("Expected vectors as a keyword argument but found None")
        vec_count =  self.index.upsert(**kwargs)
        print(f"Added ${vec_count} vectors to your index")

    def query(self, **kwargs):
        return self.index.query(**kwargs)


class WeaviateDB(VectorDB):

    def __init__(self, api_key, cluster_url):
        self.client = weaviate.Client(url=cluster_url, auth_client_secret = weaviate.AuthApiKey(api_key=api_key), additional_headers={
        "X-OpenAI-Api-Key": os.environ.get("LLM_VM_OPENAI_API_KEY")
    })

    def create_index(self, class_name, class_properties=[], distance="cosine", emb_model="text2vec-openai", mod_config=None):
        assert type(class_properties) == "list"
        if mod_config is not None:
            assert type(mod_config) == "dict" 
        class_obj = {
            'class': class_name,
            'properties': [{'name': p, 'dataType': ['text']} for p in class_properties],
            "vectorizer": emb_model,
            'moduleConfig': mod_config
        }
        self.client.schema.create_class(class_obj)
        print(f"New index {class_name} created")

    def list_indexes(self):
        schemas = self.client.schema.get()
        return schemas['classes']
    
    def describe_index(self, class_name):
        return self.client.schema.get(class_name)
    
    def delete_index(self, class_name):
        self.client.schema.delete_class(class_name)
    
    def upsert(self, class_name, batch_size=50, num_workers=1, dynamic=True, dataset=[]):
        self.client.batch.configure(batch_size=batch_size, num_workers=num_workers, dynamic=dynamic)
        with self.client.batch as batch: 
            for d in dataset:
                batch.add_data_object(d, class_name)

    def query(self, prompt, class_name, top_k=3, properties=[]):
        return self.client.query.get(class_name, properties).with_near_text({"concepts": [prompt]}).with_limit(top_k).do()

    def read_object(self, class_name, obj_id):
        return self.client.data_object.get_by_id(obj_id, class_name=class_name)
    
    def read_all_objects(self, class_name):
        collection = self.client.collections.get(class_name)
        items_list = []
        for item in collection.iterator():
            items_list.append(item.properties)
        return items_list

    def add_prop(self, class_name, prop):
        assert type(prop) == "dict"
        self.client.schema.property.create(class_name, prop)     



"""
This file contains a list of common functions across
the different agents that format and print output to the CLI.
"""

import sys


def print_big(data, label = ""):
    def do_format(x) -> str:
        formatted_title = "======#====== {:20} ======#======\n"
        if len(x) >= 20:
            return formatted_title.format(x)
        else:
            return formatted_title.format((int((20 - len(x)) / 2) * " ") + x)
    try:
        if len(label):
            print(do_format(str(label).upper()), data, flush=True, file=sys.stderr)
        else:
            print(do_format(str(data)), flush=True, file=sys.stderr)

    except:
        print(label, flush=True, file=sys.stderr)



from llm_vm.guided_completion import GrammarCompletion
from transformers import AutoTokenizer

if __name__ == '__main__':

    tokenizer = AutoTokenizer.from_pretrained("gpt2-medium", padding_side='left')
    model = GrammarCompletion("gpt2-medium", tokenizer)

    # For python
    prompt = "rand_num = 5"
    result = model.complete(prompt, temperature=1.0, max_new_tokens=20)
    print(result)

    # for JSON
    # prompt = "{ \"name\": \"Vik\", \"age\": \"50\","
    # result = model.complete(prompt, grammar_type="json", temperature=1.0, max_new_tokens=20)
    # print(result)



import sys
import openai

def delete_model (model_name: str) -> None:
    ## DELETE MODELS
    try:
        # model = openai.Model.retrieve()
        openai.Model.delete(model_name)
    except Exception as e:
        print("Could not delete model because: " + str(e), file=sys.stderr)


    try:
        # Delete all files:
        files = openai.File.list()["data"]
        [openai.File.delete(file["id"]) for file in files];
    except Exception as e:
        print("Could not delete file because: " + str(e), file=sys.stderr)



import json
from llm_vm.agents.FLAT.agent_helper.tool_utils import create_memory_prompt, make_tool_input_case
from llm_vm.agents.FLAT.agent_helper.labels import *
from typings_llm import *
from random import shuffle
from llm_vm.agents.FLAT.models.utils.tool_input_model.tool_input_model_data import tool_input_data

def __get_tool_input_jsonl(data: List[ToolInputModelData]) -> List[PromptModelEntry]:
    jsonl_entries: List[PromptModelEntry] = []

    for entry in data:
        prompt, stopper = make_tool_input_case(
            entry["mem"],
            entry["question"],
            None,
            entry["description"],
            entry["params"],
        )

        jsonl_entries.append({
            "prompt": prompt,
            "completion": json.dumps(entry["answer"]) + stopper
        })
    shuffled_examples = jsonl_entries.copy()
    shuffle(shuffled_examples)
    return shuffled_examples


def tool_input_jsonl() -> Tuple[PromptModelEntry, str]:
    return (
        __get_tool_input_jsonl(tool_input_data["data"]),
        tool_input_data["openai_model"].value
    )



import psutil
import sys

class RAMLogger:
    AMBER = '\033[93m'
    END_COLOR = '\033[0m'

    def __init__(self):
        self.start_ram = None
        self.end_ram = None

    def start(self):
        self.start_ram = self.get_ram_usage()
        print(f"{self.AMBER} RAM Usage:{self.END_COLOR}", file=sys.stderr)
        self.print_progress_bar(self.start_ram)

    def end(self):
        self.end_ram = self.get_ram_usage()
        print(f"{self.AMBER}Ending RAM Usage:{self.END_COLOR}", file=sys.stderr)
        self.print_progress_bar(self.end_ram)

    @staticmethod
    def get_ram_usage():
        return psutil.virtual_memory().percent

    def print_progress_bar(self, percentage):
        bar_length = 50
        block = int(round(bar_length * percentage / 100))
        progress = "|" + "█" * block + "-" * (bar_length - block) + "|"
        print(f"{self.AMBER}{progress} {percentage}%{self.END_COLOR}", file=sys.stderr)



import openai
import re


def tool_picker(tools_list, question, starting_tool_num):

    tools=""
    prompt= '''
{tools}
Which tool (number only), if any, would you use to answer the following question:

{question}
    '''
    count=0
    tools_list =  tools_list[starting_tool_num:]
    for i in tools_list:
        tools+="tool "+str(count)+": "+str(i["description"])+"\n"
        count+=1
    tools+="tool "+str(count)+": "+str("Use this tool when the question can be answered without using any tool or if the question is a greeting or a casual conversation. Also if we need to convert languages other than english, use this tool.")+"\n"
    prompt=prompt.format(**{"tools":tools,"question":question})
    prompt += "\n\nYOU JUST ANSWER WITH A NUMBER."
    ans = openai.Completion.create(
        model="text-davinci-003",
        max_tokens=256,
        stop=None,
        prompt=prompt,
        temperature=0.4
    )
    
    try:
        return (calcCost(prompt),str(int(re.sub("[^0-9]", "",ans['choices'][0]['text'].strip()))+starting_tool_num))
    except:
        return  (calcCost(prompt),1+starting_tool_num)

def calcCost(p):
            return (len(p) / 2700.0) * 0.02

def question_split(question,tools_list,memory):
    tools=""
    starting_tool_num=2
    count=starting_tool_num
    for i in range(len(tools_list)-starting_tool_num):
        tools+="tool "+str(i+starting_tool_num)+": "+str(tools_list[starting_tool_num+i]["description"])+"\n"
        count+=1
    tools+="tool "+str(count)+": "+str("Use this tool when the question can be answered without using any tool or if the question is a greeting or a casual conversation.")+"\n"

    prompt='''
Tools we have access to =
tool 1: The tool returns the results of free-form queries similar to those used for wolfram alpha. This is useful for complicated math or live data retrieval.  Can be used to get the current date.
tool 2: Find the driving distance and time to travel between two cities.
tool 3: Find the weather at a location and returns it in celcius.

Q="Is the distance between London and Paris larger than the distance between Boston and LA?"
Look at the tools we have access to. Split Q into subquestions to answer Q that can each be solved with one use of one tool. Make as few subquestions as possible. Split each subquestion with a comma and have no extra information other than the subquestions.
What is the distance from London to Paris?, What is the distance from Boston to LA?

Tools we have access to =
tool 1: The tool returns the results of free-form queries similar to those used for wolfram alpha. This is useful for complicated math or live data retrieval.  Can be used to get the current date.
tool 2: Find the driving distance and time to travel between two cities.
tool 3: Find the weather at a location and returns it in celcius.

Q="Who was the maternal grandfather of george washington?"
Look at the tools we have access to. Split Q into subquestions to answer Q that can each be solved with one use of one tool. Make as few subquestions as possible. Split each subquestion with a comma and have no extra information other than the subquestions.
Who was george washington's mother?,Who was her father?

Tools we have access to =
tool 1: The tool returns the results of free-form queries similar to those used for wolfram alpha. This is useful for complicated math or live data retrieval.  Can be used to get the current date.
tool 2: Find the driving distance and time to travel between two cities.
tool 3: Find the weather at a location and returns it in celcius.

Q:"What is the currency in India"
Look at the tools we have access to. Split Q into subquestions to answer Q that can each be solved with one use of one tool. Make as few subquestions as possible. Split each subquestion with a comma and have no extra information other than the subquestions.


Tools we have access to =
tool 1: The tool returns the results of free-form queries similar to those used for wolfram alpha. This is useful for complicated math or live data retrieval.  Can be used to get the current date.
tool 2: Find the driving distance and time to travel between two cities.
tool 3: Find the weather at a location and returns it in celcius.

Q:"If I am in Miami how far am I from Atlanta?"
Look at the tools we have access to. Split Q into subquestions to answer Q that can each be solved with one use of one tool. Make as few subquestions as possible. Split each subquestion with a comma and have no extra information other than the subquestions.


Tools we have access to =
{tools}
Q= "{question}"
Look at the tools we have access to. Split Q into subquestions to answer Q that can each be solved with one use of one tool. Make as few subquestions as possible. Split each subquestion with a comma and have no extra information other than the subquestions.
    '''
    prompt=prompt.format(**{"question":question,"tools":tools,"memory":memory})
    ans = openai.Completion.create(
    model="text-davinci-003",
    max_tokens=256,
    stop=None,
    prompt=prompt,
    temperature=0.2
    )
    ans=ans['choices'][0]['text'].replace("\n","").split(",")
    return (calcCost(prompt),ans)


def memory_check(memory,question):

    prompt='''
Q: "What's up?"

Is the answer to Q found in the memory or in your knowledge base already? Answer with a yes or no. yes

Q: "What color is the sky"

Is the answer to Q found in the memory or in your knowledge base already? Answer with a yes or no. yes

Q: "What is the temperature in Portland?"

Is the answer to Q found in the memory or in your knowledge base already? Answer with a yes or no. no

Memory:
{memory}

Q: "{question}"

Is the answer to Q found in the memory or in your knowledge base already? Answer with a yes or no.
    '''
    prompt=prompt.format(**{"memory":memory,"question":question})
    ans = openai.Completion.create(
                model="text-davinci-003",
                max_tokens=256,
                stop=None,
                prompt=prompt,
                temperature=0.4
            )

    ans=ans['choices'][0]['text']

    if "yes" in ans.lower():
        return (calcCost(prompt),True)
    else:
        return (calcCost(prompt),False)

def replace_variables_for_values(my_dict: dict, dynamic_keys, ignore_key: str = "_______"):
    replaced_dict = {}
    for key, value in my_dict.items():
        if (key == ignore_key):
            continue;
        formatted_key = key.format(**dynamic_keys)
        if (isinstance(value, dict)):
            formatted_value = replace_variables_for_values(value, dynamic_keys)
        elif (isinstance(value, list)):
            formatted_value = []
            for item in value:
                formatted_value += replace_variables_for_values(item, dynamic_keys)
        else:
            try:
                formatted_value = value.format(**dynamic_keys)
            except:
                formatted_value = value
        replaced_dict[formatted_key] = formatted_value
    return replaced_dict
#print(question_split(""))



import json
from llm_vm.utils.typings_llm import *
import os
import time

def persist_models(models: LLModels, openai_key: str, is_test: bool = False) -> None:
    if is_test:
        return

    current_dir = os.path.dirname(os.path.abspath(__file__))
    file_name = os.path.join(current_dir, '../models.json')

    model_json: List[LLModels] = []
    with open(file_name, "r") as output_file:
        loaded_json = json.loads(output_file.read() or "[]")
        model_json: List[LLModels] = loaded_json if loaded_json else []

    with open(file_name, "w") as output_file:
        models["TIMESTAMP"] = int(time.time())
        models["DATE"] = time.asctime()
        models["OPENAI_KEY"] = openai_key
        model_json.append(models)
        json.dump(model_json, output_file, indent=4)



import json
import openai
from agent import Agent
import time
import os, sys

google_tool = {
               'description': "Returns the result of a google search result about information on the internet. Good for retrieving information about the world and live information.",
               'dynamic_params': {"q": 'The natural language input query'},
               'method': 'GET',
               'args': {'url': "https://www.googleapis.com/customsearch/v1",
                         'params': {'key': "Enter Key Here",
                                    'cx' : 'Enter CX Here',
                                    'q': '{q}'}
                        }
               }

class suppress_output:
    def __init__(self, suppress_stdout=False, suppress_stderr=False):
        self.suppress_stdout = suppress_stdout
        self.suppress_stderr = suppress_stderr
        self._stdout = None
        self._stderr = None

    def __enter__(self):
        devnull = open(os.devnull, "w")
        if self.suppress_stdout:
            self._stdout = sys.stdout
            sys.stdout = devnull

        if self.suppress_stderr:
            self._stderr = sys.stderr
            sys.stderr = devnull

    def __exit__(self, *args):
        if self.suppress_stdout:
            sys.stdout = self._stdout
        if self.suppress_stderr:
            sys.stderr = self._stderr

openai.api_key = ""
agent = Agent(openai.api_key,[], verbose=-1)
agent.set_tools([google_tool])

f = open("compositional_celebrities.json")
data = json.load(f)
category = []
q_and_a={}
for i in data["data"]:
    if i["category"] in ['birthplace_rounded_lat', 'birthplace_rounded_lng', 'birthplace_tld', 'birthplace_ccn3', 'birthplace_currency', 'birthplace_currency_short', 'birthplace_currency_symbol', 'birthplace_jpn_common_name', 'birthplace_spa_common_name', 'birthplace_rus_common_name', 'birthplace_est_common_name', 'birthplace_urd_common_name', 'birthplace_callingcode', 'birthyear_nobelLiterature', 'birthdate_uspresident', 'birthyear_masterchamp']:
        if i["category"] not in category:
            category.append(i["category"])
            q_and_a.update({i["category"]:[(i["Question"],i["Answer"])]})
        else:
            q_and_a[i["category"]].append((i["Question"],i["Answer"]))

op = ['birthplace_rounded_lat', 'birthplace_rounded_lng', 'birthplace_tld', 'birthplace_ccn3', 'birthplace_currency', 'birthplace_currency_short', 'birthplace_currency_symbol', 'birthplace_jpn_common_name', 'birthplace_spa_common_name', 'birthplace_rus_common_name', 'birthplace_est_common_name', 'birthplace_urd_common_name', 'birthplace_callingcode', 'birthyear_nobelLiterature', 'birthdate_uspresident', 'birthyear_masterchamp']
for j in range(0,len(op)):
    print(j,op[j], file=sys.stderr)
idx=int(input())

choice=op[idx]
for i in range(100):
    with suppress_output(suppress_stdout=True, suppress_stderr=True):
        try:
            a=agent.run(q_and_a[choice][i][0],[])
        except Exception as e:
            a = [e,""]
    print(str(i)+"|",str(a[0])+"|",q_and_a[choice][i][1], file=sys.stderr)



# import llm_vm.client as l
import os, requests, json
import sys

openai_key = os.getenv('LLM_VM_OPENAI_API_KEY')

url = "http://localhost:3002/v1/complete"
json_payload = {"prompt": "what is the economic situation in canada?",
                "context": "",
                "temperature": 0.0,
                "openai_key": openai_key,
                # "finetune": True,
                }

response = requests.post(url, data=json.dumps(json_payload))
print(response.text, file=sys.stderr)



import sys
import openai
from llm_vm.utils.typings_llm import *

def check_model_status (
    job_id: str,
    label: str
) -> Dict:
    finetuned_model = openai.FineTune.retrieve(id=job_id)

    print(
        label,
        finetuned_model["status"],
        flush=True, file=sys.stderr
    )

    return finetuned_model



import json
from llm_vm.utils.typings_llm import *
from llm_vm.agents.FLAT.agent_helper.utils import *
from llm_vm.agents.FLAT.agent_helper.bothandler import question_split, pick_tool, check_can_answer_from_memory, generate_convo_history, prompt_for_answer, get_newest_decision_model
from llm_vm.agents.FLAT.agent_helper.use_tool import use_tool
from llm_vm.agents.FLAT.agent_helper.tool_utils import make_tool_input_case

# Get the current file's directory to grab the python files with common functionality in the utils/ folder
# current_dir = os.path.dirname(os.path.abspath(__file__))
# grandparent_dir = os.path.dirname(os.path.dirname(os.path.dirname(current_dir)))
# utils_dir = os.path.join(grandparent_dir, 'utils/')
# sys.path.append(utils_dir)

from labels import *
from llm_vm.agents.FLAT.agent_helper.requests.call_llm import call_llm

__MAX_RETRIES_GUESS_INPUT = 3


def __get_tool_input(
    tool: SingleTool,
    mem: TupleList,
    question: str,
    verbose: int,
    max_tokens: int = 20,
    temperature: float = 0.0
) -> Tuple[dict, float]:

    gpt_prompt, gpt_prompt_stop = make_tool_input_case(
        mem,
        question,
        None,
        tool_descr=tool["description"],
        tool_params=tool["dynamic_params"],
        wrapper_tag=L_INTERACTION
    )

    if verbose > 3:
        print_big(gpt_prompt, f"GPT PROMPT [tool={tool['id']}]")

    gpt_suggested_input, price = call_llm({
        "llm": LLMCallType.OPENAI_COMPLETION,
        "model": get_newest_decision_model(DecisionStep.INPUT),
        "max_tokens": max_tokens,
        "prompt": gpt_prompt,
        "stop": gpt_prompt_stop,
        "temperature": temperature,
    })

    try:
        if gpt_suggested_input[0] != "{":
            gpt_suggested_input = "{" + gpt_suggested_input
        if gpt_suggested_input[-1] != "}":
            gpt_suggested_input += "}"

        parsed_gpt_suggested_input: dict = json.loads(gpt_suggested_input)
        if verbose > 1:
            _has_keys = len(parsed_gpt_suggested_input.keys()) > 0
            print_big(
                parsed_gpt_suggested_input if _has_keys else "(( no input ))", "GPT SUGGESTED INPUT")
        return parsed_gpt_suggested_input, price
    except:
        print_big(gpt_suggested_input, "INVALID GPT INPUT")
        # {"$ PARSING FAILED $": str(e), "$ REASON $": "malformed JSON"}
        return {}, price


def promptf(
    question: str,
    factual_memory: TupleList,
    tools: ToolList,
    bot_instructions: str,
    verbose: int,
    split_allowed: bool = True,
    force_answer_from_memory: bool = False
) -> Tuple[str, TupleList, List[str], DebugCallList, float]:
    price_accumulator = 0

    if split_allowed:
        conversation_memory: TupleList = []
        sub_questions, main_question, price = question_split({
            "mem": factual_memory,
            "question": question,
            # "tools": tools
        })
        price_accumulator += price

        if verbose > 0:
            if len(sub_questions) == 1:
                print_big(f"NO SPLIT REQUIRED")
            else:
                print_big(f'''Question:\n- {main_question}\n\nSubquestions:\n''' + (
                    "- " + "\n- ".join(sub_questions)), f"SPLIT")

        calls: List[str] = []
        debug_return: DebugCallList = []

        for i in range(len(sub_questions)):
            answer, new_facts, new_calls, new_debug_return, price = promptf(
                sub_questions[i],
                factual_memory,
                tools,
                bot_instructions,
                verbose,
                split_allowed=False
            )
            calls = calls + new_calls
            factual_memory = factual_memory + new_facts
            debug_return = debug_return + new_debug_return
            price_accumulator += price
            conversation_memory.append((sub_questions[i], answer))

        if main_question:
            print_big(main_question, "TRYING TO ANSWER MAIN QUESTION")
            final_answer, _, _, _, price = promptf(
                main_question,
                # last round, we give the non-verbose convo memory.
                conversation_memory,
                [],
                bot_instructions,
                verbose,
                split_allowed=False,
                force_answer_from_memory=True
            )
            price_accumulator += price
            conversation_memory.append((main_question, final_answer))

            answer = final_answer
            print_big(answer, "FINAL ANSWER")
            print_big("=========#####=========")

        return answer, [], calls, debug_return, price_accumulator

    if len(tools):
        best_tool_id, price_picker = pick_tool(tools, question, factual_memory)
        price_accumulator += price_picker
        cannot_use_tools_to_answer = best_tool_id in [
            DefaultTools.ANSWER_FROM_MEMORY.value, DefaultTools.I_DONT_KNOW.value]
    else:
        best_tool_id = DefaultTools.ANSWER_FROM_MEMORY.value if force_answer_from_memory else DefaultTools.I_DONT_KNOW.value
        cannot_use_tools_to_answer = True

    # Only in the cases where we DON'T have a tool (answer_from_tool == False) we check if we can answer from memory.
    can_answer_from_memory = force_answer_from_memory
    if cannot_use_tools_to_answer and not force_answer_from_memory:
        can_answer_from_memory, price_check = check_can_answer_from_memory(
            question, memory=factual_memory)
        price_accumulator += price_check

    print_big(
        f"MEM ({can_answer_from_memory}) / TOOL ({not cannot_use_tools_to_answer})")

    if cannot_use_tools_to_answer and can_answer_from_memory:

        cur_prompt = (
            bot_instructions +
            generate_convo_history(memory=factual_memory) +
            f"\n\n" +
            f"\n[SYSTEM]: Use <{L_ANSWER_DATA}/> to answer better, but follow the format of <{L_ANSWER_SUMMARY}/>\n" +
            prompt_for_answer(question)
        )

        answer, price = call_llm({
            "llm": LLMCallType.OPENAI_COMPLETION,
            "model": OpenAIModel.DAVINCI_TEXT.value if "quality" == "best" else OpenAIModel.CURIE_TEXT.value,
            "max_tokens": 200,
            "prompt": cur_prompt,
            "stop": None,
            "temperature": 0.1,
        })

        price_accumulator += price
        if verbose > 0:
            print_op("Question:  ", question)
            print_op("AI Answer: ", answer)

        return answer, [(question, answer)], [], [], price_accumulator

    conversation_history = generate_convo_history(facts=factual_memory)
    # No information in the Conversation memory AND no tool --> Using AI General Knowledge Base.
    if best_tool_id in [DefaultTools.I_DONT_KNOW.value, DefaultTools.ANSWER_FROM_MEMORY.value]:

        if (best_tool_id == DefaultTools.I_DONT_KNOW.value):
            current_prompt = f"{bot_instructions}\n" if bot_instructions else "" \
                + (f"<{L_CONVERSATION}>{conversation_history}</{L_CONVERSATION}>\n" if len(conversation_history) else "") \
                + f"<{L_QUESTION}>{question}</{L_QUESTION}>\n" \
                + f"<{L_THOUGHT}>Maybe not enough information to answer. If that's the case, say why.</{L_THOUGHT}>\n" \
                + f"<{L_ANSWER}>"
        else:
            current_prompt = f"{bot_instructions}\n" if bot_instructions else "" \
                + (f"<{L_CONVERSATION}>{conversation_history}</{L_CONVERSATION}>\n" if len(conversation_history) else "") \
                + f"<{L_QUESTION}>{question}</{L_QUESTION}>\n" \
                + f"<{L_ANSWER}>"

        a, price = call_llm({
            "llm": LLMCallType.OPENAI_COMPLETION,
            "model": OpenAIModel.DAVINCI_TEXT.value,
            "max_tokens": 500,
            "prompt": current_prompt,
            "stop": f"</{L_ANSWER}>",
            "temperature": 0.1,
        })
        price_accumulator += price
        return a, [(question, a)], [], [], price_accumulator

    best_tool: SingleTool = get_tool_by_id(tools, best_tool_id)

    if "dynamic_params" not in best_tool:
        best_tool["dynamic_params"] = {}

    dynamic_params_to_fill = len(best_tool["dynamic_params"].keys())
    # only try to guess the "dynamic_params" if we have any
    if dynamic_params_to_fill == 0:
        tool_input = {}
    else:
        # Try more than once to get the right input.
        for attempt_to_find_input in range(__MAX_RETRIES_GUESS_INPUT):
            tool_input, price = __get_tool_input(
                best_tool,
                factual_memory,
                question,
                verbose,
                max_tokens=200,
                # higher temperature on subsequent attempts
                temperature=min(0.15 * attempt_to_find_input, 0.5)
            )
            price_accumulator += price

            filled_dynamic_params = len(tool_input.keys())
            if filled_dynamic_params == dynamic_params_to_fill:
                break

        # If, besides all attempts, we still don't have the right input...
        if filled_dynamic_params != dynamic_params_to_fill:
            raise Exception(
                f"Wrong input! Not enough parameters (actual: {filled_dynamic_params}, expected: {dynamic_params_to_fill})")

    try:

        response_instructions = question

        if "ai_response_prompt" in best_tool.keys():
            response_instructions = best_tool["ai_response_prompt"]

        answer, actual_url, raw_api_return, price = use_tool(
            best_tool,
            tool_input,
            question,
            factual_memory,
            verbose,
            bot_instructions,
            response_instructions
        )
        price_accumulator += price
    except Exception as e:
        print_big(e, "EXCEPTION CAUGHT! [#1]")
        cur_prompt_text = f"\n"\
            f"{bot_instructions}\n" \
            f"{conversation_history}\n" \
            f"--------\n" + \
            prompt_for_answer(question)

        a, price = call_llm({
            "llm": LLMCallType.OPENAI_COMPLETION,
            "model": OpenAIModel.DAVINCI_TEXT.value,
            "max_tokens": 500,
            "prompt": cur_prompt_text,
        })
        price_accumulator += price
        return a, [(question, a)], [], [], price_accumulator

    if verbose > 0:
        print_big("RESULT OF ANSWER")
        print_op("Question:  ", question)
        print_op("Tool Input:", tool_input)
        print_op("AI Answer: ", answer)

    return answer, \
        [(question, verbose_answer(raw_api_return, answer))], \
        [actual_url], \
        [{"response": "", "gpt_suggested_params": tool_input, "url": actual_url, "raw_api_return": raw_api_return}], \
        price_accumulator



import re
import openai
import random
from llm_vm.utils.typings_llm import *
from llm_vm.agents.FLAT.agent_helper.requests.call_llm import call_llm

from llm_vm.agents.FLAT.agent_helper.tools import *
from llm_vm.agents.FLAT.agent_helper.tool_utils import *
from llm_vm.agents.FLAT.agent_helper.utils import *

from llm_vm.agents.FLAT.models.get_decision_model import get_newest_decision_model
from llm_vm.agents.FLAT.models.utils.tool_picker_model.tool_picker_model_data import tool_input_data
from llm_vm.agents.FLAT.models.utils.question_split_model.question_split_model_data import question_splitter_data

# openai.api_key = OPENAI_DEFAULT_KEY


def question_split(question_to_split, use_fine_tuned_model = True):

    question_data = question_splitter_data["data"].copy()
    random.shuffle(question_data)
    NUMBER_OF_EXAMPLES = 6
    question_data = question_data[:NUMBER_OF_EXAMPLES]
    use_fine_tuned_model = False
    prompt, prompt_stop = splitter_prompt(
        question_data +
        [question_to_split]
    )

    if use_fine_tuned_model:
            custom_model, custom_model_key = get_newest_decision_model(DecisionStep.SPLIT)
    else:
        custom_model, custom_model_key = OpenAIModel.DAVINCI_TEXT.value, None

    sub_questions_str, price = call_llm({
        "llm": LLMCallType.OPENAI_COMPLETION,
        "model": (custom_model, custom_model_key),
        "max_tokens": 256,
        "stop": prompt_stop,
        "prompt": prompt,
        "temperature": 0.2,
    })

    sub_questions, main_question = tidy_up_subquestions(
        sub_questions_str.replace(prompt_stop, ""),
        question_to_split["question"]
    )

    return sub_questions, main_question, price


def pick_tool(tools_list, question, conversation_history, use_fine_tuned_model = True, debug_prompt = False):
    tool_data = tool_input_data["data"].copy()
    random.shuffle(tool_data)
    NUMBER_OF_EXAMPLES = 6
    tool_data = tool_data[:NUMBER_OF_EXAMPLES]
    use_fine_tuned_model = False


    prompt, stop = toolpicker_prompt(
        tool_data + [{"question": question, "mem": conversation_history}],
        tools_list,
    )

    if debug_prompt:
        print_big(prompt, "toolpicker prompt")

    if use_fine_tuned_model:
        custom_model, custom_model_key = get_newest_decision_model(
            DecisionStep.TOOL_PICKER)
    else:
        custom_model, custom_model_key = OpenAIModel.DAVINCI_TEXT.value, None

    suggested_tool_str, price = call_llm({
        "llm": LLMCallType.OPENAI_COMPLETION,
        "model": (custom_model, custom_model_key),
        "max_tokens": 10,
        "stop": stop,
        "prompt": prompt,
        "temperature": 0,
    })

    number_match_regex = "-?[0-9]+"

    try:
        valid_tool_ids = [t["id"] for t in tools_list] + \
            [DefaultTools.ANSWER_FROM_MEMORY.value, DefaultTools.I_DONT_KNOW.value]
        parsed_suggested_tool_str = re.findall(
            number_match_regex, suggested_tool_str)
        if len(parsed_suggested_tool_str) == 0:
            raise Exception("No matching ID found in response")

        suggested_tool_id = int(parsed_suggested_tool_str[0])
        if suggested_tool_id not in valid_tool_ids:
            raise Exception(
                f"Invalid tool id: {suggested_tool_id} // {','.join([str(v) for v in valid_tool_ids])}")

        tools_list.sort(key=lambda t: t["id"])
        print_big(
            "".join([f'{"✓" if tool["id"] == suggested_tool_id else " "} {tool["id"]}.- {tool["description"]}\n' for tool in tools_list]) +
            f"\n> Question: {question}\n> Raw answer: '{suggested_tool_str}'\n> Tool ID: {suggested_tool_id}",
            "LIST OF DATA TOOLS"
        )

        return suggested_tool_id, price
    except Exception as e:
        print_big(f'Exception parsing tool_id: "{suggested_tool_str}"')
        print(e, flush=True, file=sys.stderr)
        return DefaultTools.I_DONT_KNOW.value, price


def check_can_answer_from_memory(question: str, memory: TupleList = [], facts: TupleList = []) -> Tuple[bool, float]:

    memory_prompt, memory_stop = create_memory_prompt(
        # answer_from_memory_data["data"] +
        [{"question": question, "mem": memory, "facts": facts}]
    )

    ans, price = call_llm({
        "llm": LLMCallType.OPENAI_COMPLETION,
        "model": get_newest_decision_model(DecisionStep.FROM_MEMORY),
        "max_tokens": 10,
        "stop": memory_stop,
        "prompt": memory_prompt,
        "temperature": 0.1,
    })

    if "yes" in ans.lower():
        return True, price
    else:
        return False, price



import requests
import traceback
import os
import random
import concurrent.futures
import openai
import urllib.request
import json
import string
import re
import sys



def flatten(a):
    return sum(a, [])


def print_op(*kargs, **kwargs):
    print(*kargs, **kwargs, flush=True, file=sys.stderr)


def prepPrintPromptContext(p):
    return "--> " + p.replace("\n", "\n--> ") if len(p) > 0 else ""


def MSG(u, c):
    return [{"role": u, "content": c}]


def call_ChatGPT(state, cur_prompt, stop=None, max_tokens=20, temperature=0.2):
    if state.verbose > 1:
        print_op("\nGPT input for {" + str(stop) + "} " + str(len(cur_prompt)) + ".")
    if state.verbose > 2:
        print_op(str(cur_prompt))

    ppt = 0.002

    def calcCost(p):
        chars = sum((len(a["content"]) for a in p))
        if state.verbose > 0:
            print_op("ASK_CHARS:", chars)
        c = (chars / 2700.0) * ppt
        if state.verbose > 2:
            print_op("PrePrice: $" + str(c))
        return c

    cost = calcCost(cur_prompt)
    try:
        ans = openai.ChatCompletion.create(
            model="gpt-3.5-turbo-0301",
            max_tokens=max_tokens,
            stop=stop,
            messages=cur_prompt,
            temperature=temperature,
        )

    except Exception as e:
        state.price += cost
        traceback.print_exc()
        print_op("Error:", e)

        return "OpenAI is down!"

    price = ppt * ans["usage"]["total_tokens"] / 1000
    if state.verbose > 0:
        print_op("Price: $" + str(price))
    state.price += price
    response_text = ans["choices"][0]["message"]["content"]

    if state.verbose > 2:
        print_op("GPT output:")
        print_op(prepPrintPromptContext(response_text))
        print_op("GPT output fin.\n")

    return response_text


def call_gpt(
    state, cur_prompt: str, stop: str, max_tokens=20, quality="best", temperature=0.0
):
    if state.verbose > 1:
        print_op("\nGPT input for {" + stop + "} " + str(len(cur_prompt)) + ".")
    if state.verbose > 2:
        print_op(prepPrintPromptContext(cur_prompt))

    ask_tokens = max_tokens + len(cur_prompt) / 2.7

    if state.verbose > 0:
        print_op("ASK_TOKENS:", ask_tokens)

    if (ask_tokens) > 2049:
        quality = "best"

    model = {
        "best": ("text-davinci-003", 0.02),
        "okay": ("text-curie-001", 0.002),
    }[quality]

    def calcCost(p):
        return (len(p) / 2700.0) * model[1]

    cost = calcCost(cur_prompt)

    try:
        ans = openai.Completion.create(
            model=model[0],
            max_tokens=max_tokens,
            stop=stop,
            prompt=cur_prompt,
            temperature=temperature,
        )
    except Exception as e:
        print_op("WTF:", e)
        state.price += cost
        return "OpenAI is down!"

    response_text = ans["choices"][0]["text"]

    simpleprice = model[1] * ans["usage"]["total_tokens"] / 1000
    if state.verbose > 0:
        print_op("SimplePrice: $" + str(simpleprice))
    state.price += simpleprice

    if state.verbose > 2:
        print_op("GPT output:")
        print_op(prepPrintPromptContext(response_text))
        print_op("GPT output fin.\n")

    return response_text


def deep_fmap(lambdaFunc, json_data):
    print_op(json_data)
    if isinstance(json_data, list):
        print_op("##LIST")
        return list(map(lambda listItem: deep_fmap(lambdaFunc, listItem), json_data))
    elif isinstance(json_data, tuple):
        print_op("##TUPLE")
        return tuple(map(lambda tupleItem: deep_fmap(lambdaFunc, tupleItem), json_data))
    elif isinstance(json_data, dict):
        print_op("##DICT")
        return {lambdaFunc(k): deep_fmap(lambdaFunc, v) for k, v in json_data.items()}
    else:
        print_op("##SIMPLE")
        return lambdaFunc(json_data)


def replace_variables_for_values(
    my_dict: dict, dynamic_keys, ignore_key: str = "_______"
):
    replaced_dict = {}
    for key, value in my_dict.items():
        if key == ignore_key:
            continue
        formatted_key = key.format(**dynamic_keys)
        if isinstance(value, dict):
            formatted_value = replace_variables_for_values(value, dynamic_keys)
        elif isinstance(value, list):
            formatted_value = []
            for item in value:
                formatted_value += replace_variables_for_values(item, dynamic_keys)
        else:
            try:
                formatted_value = value.format(**dynamic_keys)
            except:
                formatted_value = value
        replaced_dict[formatted_key] = formatted_value
    return replaced_dict


def tool_api_call(self, tool, gpt_suggested_input, question, memory, facts, query=""):
    if query == "":
        query = question

    if gpt_suggested_input[0] != "{":
        gpt_suggested_input = "{" + gpt_suggested_input
    if gpt_suggested_input[-1] != "}":
        gpt_suggested_input += "}"

    if self.verbose > 1:
        print_op("GPT SUGGESTED INPUT:", gpt_suggested_input)

    parsed_gpt_suggested_input = json.loads(gpt_suggested_input)
    # make sure all of the suggested fields exist in the tool desc
    for i in parsed_gpt_suggested_input.keys():
        if i not in tool["dynamic_params"].keys():
            raise Exception("Bad Generated Input")
    tool_args = replace_variables_for_values(tool["args"], parsed_gpt_suggested_input)

    url = tool_args["url"]
    if self.verbose > -1:
        print_op(tool["method"] + ":", url)

    if "auth" in tool_args and isinstance(tool_args["auth"], dict):
        auths = list(tool_args["auth"].items())
        if len(auths) > 0:
            tool_args["auth"] = list(tool_args["auth"].items())[0]
        else:
            del tool_args["auth"]

    # Remove those parameters that are not part of Python's `requests` function.
    tool_args.pop("jsonParams", None)
    tool_args.pop("urlParams", None)

    if self.verbose > -1:
        print_op("ARGS: ", tool_args)

    resp = (requests.get if tool["method"] == "GET" else requests.post)(**tool_args)

    # print_op("FINAL URL: (" + tool["method"] + ") ", resp.url)

    actual_call = str(tool_args)

    if resp.status_code in [404, 401, 500]:
        er = " => " + str(resp.status_code)
        return "This tool isn't working currently" + er

    ret = str(resp.text)
    if self.verbose > 4:
        print(ret, file=sys.stderr)
    try:
        ret = str(json.loads(ret))
    except:
        pass

    if len(ret) > 10000:
        ret = ret[0:10000]
    mem = "".join(
        [
            self.makeInteraction(p, a, "P", "AI", INTERACTION="Human-AI")
            for p, a in memory
        ]
    ) + "".join(
        [self.makeInteraction(p, a, "P", "AI", INTERACTION="AI-AI") for p, a in facts]
    )

    prompt = MSG("system", "You are a good and helpful bot" + self.bot_str)
    prompt += MSG(
        "user",
        mem
        + "\nQ:"
        + query
        + "\n An api call about Q returned:\n"
        + ret
        + "\nUsing this information, what is the answer to Q?",
    )
    a = call_ChatGPT(self, prompt, stop="</AI>", max_tokens=256).strip()
    return a



from llm_vm.utils.typings_llm import *
from llm_vm.agents.FLAT.agent_helper.labels import *
from llm_vm.agents.FLAT.agent_helper.tool_utils import toolpicker_prompt
from llm_vm.agents.FLAT.models.utils.tool_picker_model.tool_picker_model_data import tool_input_data

def __construct_tool_picker_jsonl(data: List[ToolpickerInputModelData]) -> List[PromptModelEntry]:
    jsonl_entries: List[PromptModelEntry] = []
    for entry in data:
        prompt, stopper = toolpicker_prompt(
            [{
                "mem": entry["mem"] if "mem" in entry else None,
                "question": entry["question"],
                "thought": entry["thought"] if "thought" in entry else None
            }],
            entry["tools"]
        )
        jsonl_entries.append({
            "prompt": prompt,
            "completion": str(entry["answer"]) + stopper
        })
    return jsonl_entries


def tool_picker_jsonl() -> Tuple[PromptModelEntry, str]:
    return (
        __construct_tool_picker_jsonl(tool_input_data["data"]),
        tool_input_data["openai_model"].value
    )



import os
import sys

# Get the current file's directory to grab the python files with common functionality in the utils/ folder
# current_dir = os.path.dirname(os.path.abspath(__file__))
# grandparent_dir = os.path.dirname(os.path.dirname(os.path.dirname(current_dir)))
# utils_dir = os.path.join(grandparent_dir, 'utils/')
# sys.path.append(utils_dir)

from keys import *
from labels import *
from llm_vm.utils.typings_llm import *
from llm_vm.agents.FLAT.agent_helper.utils import verbose_answer

CUSTOM_TOOL_ANSWER_EMBEDDING = "/answer_embedding"


def __get_generic_tools():
    # # wolfram

    WOLFRAM_KEY = os.getenv("LLM_VM_WOLFRAM_KEY")
    GOOGLE_MAPS_KEY = os.getenv("LLM_VM_GOOGLE_MAPS_KEY")
    wolfram_tool = {
        'description': "Useful to query questions about people, events, anything that can change, complicated math, live data retrieval, current date and other data.",
        # {'description': "The tool returns the results of free-form queries similar to those used for wolfram alpha. This is useful for complicated math or live data retrieval.  Can be used to get the current date.",
        'id': DefaultTools.WOLFRAM.value,
        'dynamic_params': {"input": 'The natural language input query'},
        'method': 'GET',
        'args': {
            'url': "http://api.wolframalpha.com/v2/query",
            'params': {'appid': WOLFRAM_KEY, 'input': '{input}'}
        },
    }

    # geopy
    directions_tool = {
        'description': "Find the driving distance and time to travel between two cities.",
        'id': DefaultTools.DIRECTIONS.value,
        'dynamic_params': {"origins": 'the origin city', "destinations": 'the destination city'},
        'method': 'GET',
        'args': {
            'url': "https://maps.googleapis.com/maps/api/distancematrix/json",
            'params': {
                'key': GOOGLE_MAPS_KEY,
                'origins': '{origins}',
                'destinations': '{destinations}'
            }
        }
    }

    # weather
    weather_tool = {
        'description': 'Useful to get the weather.',
        'ai_response_prompt': 'Returns temperature in celcius',
        'id': DefaultTools.WEATHER.value,
        'dynamic_params': {
            "latitude": 'latitude of the location, a float',
            "longitude": 'the longitude of the location, a float'
        },
        "ai_response_prompt": '''the value of "current_weather.weathercode" means:
0: Clear sky
1, 2, 3: Mainly clear, partly cloudy, and overcast
45, 48: Fog and depositing rime fog
51, 53, 55: Drizzle
56, 57: Freezing Drizzle
61, 63, 65: Rain
66, 67: Freezing Rain
71, 73, 75: Snow fall
77: Snow grains
80, 81, 82: Rain showers
85, 86: Snow showers slight and heavy
95: Thunderstorm
96, 99: Thunderstorm with slight and heavy hail

Do not return the weathercode as a number, instead use the description from list above''',
        'method': 'GET',
        'args': {
            'url': "https://api.open-meteo.com/v1/forecast",
            'params': {
                'current_weather': 'true',
                'latitude': '{latitude}',
                'longitude': '{longitude}'
            }
        }
    }

    # return [wolfram_tool, directions_tool, weather_tool]
    return [weather_tool]


GENERIC_TOOLS = __get_generic_tools()
